from typing import Dict, List, Type, TYPE_CHECKING

from cosmoglobe.sky import COSMOGLOBE_COMPS

if TYPE_CHECKING:
    from cosmoglobe.h5.chain_context import ChainContext


class ChainContextFactory:
    """Factory that book-keeps and registeres contexts for components."""

    def __init__(self):
        self._context: Dict[str, List["ChainContext"]] = {
            component: [] for component in COSMOGLOBE_COMPS
        }
        self._mappings: Dict[str, Dict[str, str]] = {
            component: {} for component in COSMOGLOBE_COMPS
        }
        self._units: Dict[str, Dict[str, str]] = {
            component: {} for component in COSMOGLOBE_COMPS
        }

    def register_context(
        self, components: List[str], context: Type["ChainContext"]
    ) -> None:
        """Registers a specific context for a set of components."""

        if not components:
            for component in self._context.keys():
                self._context[component].append(context())
            return

        for component in components:
            if component not in self._context:
                raise ValueError(
                    "Cannot register a context for a non cosmoglobe component"
                )
            self._context[component].append(context())

    def register_mapping(self, components: List[str], mapping: Dict[str, str]) -> None:
        """Registers a specific mapping between chain and sky model notation."""

        if not components:
            for component in self._mappings.keys():
                self._mappings[component].update(mapping)
            return

        for component in components:
            if component not in self._mappings:
                raise ValueError(
                    "Cannot register a mapping for a non cosmoglobe component"
                )
            self._mappings[component].update(mapping)

    def register_units(self, components: List[str], units: Dict[str, str]) -> None:
        """Registers a units for values from the chain."""

        if not components:
            for component in self._units.keys():
                self._units[component].update(units)
            return

        for component in components:
            if component not in self._units:
                raise ValueError(
                    "Cannot register a unit for a non cosmoglobe component"
                )
            self._units[component].update(units)

    def get_context(self, component: str) -> List["ChainContext"]:
        """Returns the context."""

        if component not in self._context:
            raise KeyError(f"No context is registered for {component=}")

        return [context for context in self._context[component]]

    def get_mappings(self, component: str) -> Dict[str, str]:
        """Returns the mappings between notation for a component."""

        if component not in self._mappings:
            raise KeyError(f"No mappings are registered for {component=}")

        return self._mappings[component]

    def get_units(self, component: str) -> Dict[str, str]:
        """Returns the mappings between notation for a component."""

        if component not in self._units:
            raise KeyError(f"No units are registered for {component=}")

        return self._units[component]



from pathlib import Path
import textwrap
from typing import Any, Dict, Generator, List, Optional, Union, overload

import h5py
import numpy as np

from cosmoglobe.h5 import ChainVersion, PARAMETER_GROUP_NAME
from cosmoglobe.h5.alms import unpack_alms_from_chain
from cosmoglobe.h5.decorators import validate_key, validate_samples, unpack_alms
from cosmoglobe.h5.exceptions import ChainFormatError, ChainSampleError
from cosmoglobe.sky import COSMOGLOBE_COMPS


class Chain:
    """An interface for Cosmoglobe chainfiles.

    This class aims to provide a convenient interface for working with
    Cosmoglobe chain files.
    """

    def __init__(self, path: Union[str, Path], burn_in: Optional[int] = None) -> None:
        """Validate and initialize the Chain object."""

        if not (path := Path(path)).is_file():
            raise FileNotFoundError(f"{path.name} was not found")
        try:
            with h5py.File(path, "r") as file:
                pass
        except OSError:
            raise ChainFormatError(f"{path.name} is not a HDF5 file")

        with h5py.File(path, "r") as file:
            samples = list(file.keys())
            if not samples:
                raise ChainFormatError("chain has no samples")
            try:
                samples.remove(PARAMETER_GROUP_NAME)
                version = ChainVersion.NEW
            except ValueError:
                version = ChainVersion.OLD

            sampled_groups = list(file[samples[0]].keys())
            components = [
                group for group in sampled_groups if group in COSMOGLOBE_COMPS
            ]

            parameters: Dict[str, Dict[str, Any]] = {}
            if version is ChainVersion.NEW:
                for component, group in file[PARAMETER_GROUP_NAME].items():
                    parameters[component] = {}
                    for key, value in group.items():
                        if np.issubdtype(value.dtype, np.string_):
                            value = value.asstr()
                        parameters[component][key] = value[()]

        if burn_in is None:
            self._samples = samples
        else:
            if burn_in >= (nsamples := len(samples)):
                raise ChainSampleError(f"{burn_in=} out of range with {nsamples=}")
            self._samples = samples[burn_in:]
        self._components = components
        self._parameters = parameters
        self._path = path
        self._version = version

    @property
    def samples(self) -> List[str]:
        """List of all samples in the chain."""

        return self._samples

    @property
    def nsamples(self) -> int:
        """Number of samples in the chain."""

        return len(self.samples)

    @property
    def components(self) -> List[str]:
        """List of the sky components in the chain."""

        return self._components

    @property
    def parameters(self) -> Dict[str, Dict[str, Any]]:
        """Dictionary of the parameters in the parameter group in the chain."""

        return self._parameters

    @property
    def path(self) -> Path:
        """Path to the chainfile."""

        return self._path

    @property
    def version(self) -> ChainVersion:
        """Chain version."""

        return self._version

    @overload
    def get(self, key: str, *, samples: Optional[range] = None) -> Any:
        ...

    @overload
    def get(self, key: str, *, samples: Optional[int] = None) -> Any:
        ...

    @validate_key
    @validate_samples
    @unpack_alms
    def get(self, key: str, *, samples: List[str]) -> Any:
        """Returns the value of an key for all samples.

        Parameters
        ----------
        key
            The path to an item that has been sampled in the chain, e.g
            'dust/amp_alm'.
        samples
            An int or a range of samples for which to return the value. If
            None, all samples in the chain are used.

        Returns
        -------
        values
            The value of the key for each samples.
        """

        with h5py.File(self.path, "r") as file:
            values = [file[f"{sample}/{key}"][()] for sample in samples]

        return np.asarray(values) if len(values) != 1 else values[0]

    @overload
    def mean(self, key: str, *, samples: Optional[range] = None) -> Any:
        ...

    @overload
    def mean(self, key: str, *, samples: Optional[int] = None) -> Any:
        ...

    @validate_key
    @validate_samples
    @unpack_alms
    def mean(self, key: str, *, samples: List[str]) -> Any:
        """Returns the mean of an key over all samples.

        Parameters
        ----------
        key
            The path to an item that has been sampled in the chain, e.g
            'dust/amp_alm'.
        samples
            An int or a range of samples to average over. If None, all
            samples in the chain are used.

        Returns
        -------
        value
            The averaged value of the key over all samples.
        """

        with h5py.File(self.path, "r") as file:
            value = file[f"{samples[0]}/{key}"][()]
            dtype = value.dtype.type
            if len(samples) > 1:
                for sample in samples[1:]:
                    value += file[f"{sample}/{key}"][()]

        return dtype(value / len(samples))  # Converting back to original dtype

    @overload
    def load(self, key: str, *, samples: Optional[range] = None) -> Any:
        ...

    @overload
    def load(self, key: str, *, samples: Optional[int] = None) -> Any:
        ...

    @validate_key
    @validate_samples
    def load(self, key: str, *, samples: List[str]) -> Generator:
        """Returns a generator to be used in a for loop.

        NOTE to devs: The unpack_alms decorator wont work on this function
        due to it not processing the returned data until it is iterated over.

        Parameters
        ----------
        key
            The path to an item that has been sampled in the chain, e.g
            'dust/amp_alm'.
        samples
            An int or a range of samples to average over. If None, all
            samples in the chain are used.

        Returns
        -------
            A generator that can be looped over to yield each sampled value.
        """

        with h5py.File(self.path, "r") as file:
            for sample in samples:
                value = file[f"{sample}/{key}"][()]
                if "alm" in key:
                    value = unpack_alms_from_chain(self, value, key)

                yield value

    @validate_key
    @unpack_alms
    def __getitem__(self, key: str) -> Any:
        """Returns the value of a key from the chain.

        Parameters
        ----------
        key
            The *full* path to an item in the chain.

        Returns
        -------
            The value of the key.
        """

        with h5py.File(self.path, "r") as file:
            item = file[key]
            if isinstance(item, h5py.Group):
                return list(item.keys())

            elif isinstance(item, h5py.Dataset):
                if np.issubdtype(item.dtype, np.string_):
                    return item.asstr()[()]
                return item[()]

    def _format_samples(self, samples: Union[List[int], int]) -> Union[List[str], str]:
        """Converts a range to the string format of the samples in the chain."""

        leading_zeros = len(self.samples[0])
        if isinstance(samples, list):
            return [f"{sample:0{leading_zeros}d}" for sample in samples]

        return f"{samples:0{leading_zeros}d}"

    def __str__(self) -> str:
        """Representation of the chain."""

        COL_LEN = 40
        CHAIN_META = {
            "Num Samples": self.nsamples,
            "Num Components": len(self.components),
            "Size": f"{self.path.stat().st_size / (1024 * 1024 * 1024):.3f}" + " GB",
        }

        def center(string: str, fill=" ") -> str:
            white_space_len = (COL_LEN - len(string)) // 2
            white_space = fill * white_space_len
            output = f"{white_space}{string}{white_space}"
            if len(output) < COL_LEN:
                output += fill
            return output

        main_repr = "\n"
        main_repr += "-" * COL_LEN + "\n"
        main_repr += center(self.path.name) + "\n"
        main_repr += "-" * COL_LEN + "\n"
        main_repr += "\n"

        for key, value in CHAIN_META.items():
            main_repr += f"{key:<{COL_LEN//2 - 1}}{'='}{value:>{COL_LEN//2}}\n"

        main_repr += "\n"
        main_repr += center(" Components ", fill="-") + "\n"
        main_repr += (
            center(textwrap.fill("  ".join(self.components), width=COL_LEN)) + "\n"
        )
        main_repr += "\n"
        main_repr += "-" * COL_LEN + "\n"

        return main_repr



import click
from rich import print
import astropy.units as u
import matplotlib.pyplot as plt
import os
import re

from cosmoglobe.plot import plot as cplot
from cosmoglobe.plot import trace as ctrace
from cosmoglobe.plot import gnom as cgnom


from .plottools import *


@click.group()
def commands_plotting():
    pass


@commands_plotting.command()
@click.argument("input", nargs=-1)
@click.option(
    "-sig",
    default=0,
    help="Signal to be plotted 0 by default (0, 1, 2 is interprated as IQU)",
)
@click.option(
    "-comp",
    default=None,
    type=click.STRING,
    help="Component name for autosetting plottingparameters",
)
@click.option(
    "-freq",
    default=None,
    type=click.FLOAT,
    help="Frequency in GHz needed for scaling maps when using a model object input",
)
@click.option(
    "-ticks",
    default=None,
    help="Min and max value for data. If None, uses 97.5th percentile.",
)
@click.option(
    "-min",
    default=None,
    type=click.FLOAT,
    help="Min value of colorbar, overrides autodetector.",
)
@click.option(
    "-max",
    default=None,
    type=click.FLOAT,
    help="Max value of colorbar, overrides autodetector.",
)
@click.option("-range", "rng", default=None, type=click.FLOAT, help="Color bar range")
@click.option(
    "-nocbar",
    is_flag=True,
    help='Adds colorbar ("cb" in filename)',
)
@click.option(
    "-unit",
    default=None,
    type=click.STRING,
    help="Set unit (Under color bar), has LaTeX functionality. Ex. $\mu$",
)
@click.option(
    "-fwhm",
    default=0.0,
    type=click.FLOAT,
    help="FWHM of smoothing, in arcmin.",
)
@click.option(
    "-nside",
    default=None,
    type=click.INT,
    help="nside for optional ud_grade.",
)
@click.option(
    "-sample",
    default=-1,
    type=click.INT,
    help="Select sample when inputing .h5 file",
)
@click.option(
    "-mask",
    default=None,
    type=click.STRING,
    help="Masks input with specified maskfile.",
)
@click.option(
    "-maskfill",
    default=None,
    type=click.STRING,
    help='Color to fill masked area. for example "gray". Transparent by default.',
)
@click.option(
    "-cmap",
    default=None,
    type=click.STRING,
    help="Chose colormap (ex. sunburst, planck, etc). Available are matplotlib and cmasher. Also qualitative plotly [ex. q-Plotly-4 (q for qualitative 4 for max color)]",
)
@click.option(
    "-norm",
    default=None,
    help='"Linear" is normal, "log" Normalizesusingsemi-logscale',
)
@click.option(
    "-remove_dip",
    is_flag=True,
    help="Fits a dipole to the map and removes it.",
)
@click.option(
    "-remove_mono",
    is_flag=True,
    help="Fits a monopole to the map and removes it.",
)
@click.option(
    "-title",
    default=None,
    type=click.STRING,
    help="Sets the full figure title. Has LaTeX functionaliity (ex. $A_{s}$.)",
)
@click.option(
    "-right_label",
    default=None,
    type=click.STRING,
    help="Sets the upper right title. Has LaTeX functionaliity (ex. $A_{s}$.)",
)
@click.option(
    "-left_label",
    default=None,
    type=click.STRING,
    help="Sets the upper left title. Has LaTeX functionaliity (ex. $A_{s}$.)",
)
@click.option(
    "-width",
    default=None,
    help="Size: in inches OR 1/3, 1/2 and full page width (8.8/12/18cm) [ex. x, s, m or l, or ex. slm for all], m by default",
)
@click.option(
    "-xsize",
    default=1000,
    type=click.INT,
    help="figuresize in px",
)
@click.option(
    "-darkmode",
    is_flag=True,
    help='Plots all outlines in white for dark bakgrounds ("dark" in filename)',
)
@click.option(
    "-interactive",
    is_flag=True,
    help="Uses healpys interactive mode, zooming is possible etc.",
)
@click.option(
    "-rot",
    default=None,
    help="Describe the rotation to apply. In the form (lon, lat, psi) (unit: degrees) : the point at longitude *lon* and latitude *lat* will be at the center. An additional rotation of angle *psi* around this direction is applied.",
)
@click.option(
    "-coord",
    default=None,
    help="Either one of 'G', 'E' or 'C' to describe the coordinate system of the map,",
)
@click.option(
    "-nest",
    is_flag=True,
    help="If True, ordering scheme is NESTED.",
)
@click.option(
    "-flip",
    default="astro",
    type=click.STRING,
    help="Defines the convention of projection : 'astro' or 'geo'",
)
@click.option(
    "-graticule",
    is_flag=True,
    help="Adds graticule",
)
@click.option(
    "-graticule_labels",
    is_flag=True,
    help="longitude and latitude labels",
)
@click.option(
    "-return_only_data",
    is_flag=True,
    help="Returns figure",
)
@click.option(
    "-projection_type",
    default="mollweide",
    type=click.STRING,
    help='{"aitoff", "hammer", "lambert", "mollweide", "cart", "3d", "polar"} type of the plot',
)
@click.option(
    "-cb_orientation",
    default="horizontal",
    type=click.STRING,
    help='{"horizontal", "vertical"} color bar orientation',
)
@click.option(
    "-xlabel",
    default=None,
    type=click.STRING,
    help="set x axis label",
)
@click.option(
    "-ylabel",
    default=None,
    type=click.STRING,
    help="set y axis label",
)
@click.option(
    "-longitude_grid_spacing",
    default=60,
    type=click.FLOAT,
    help="set x axis grid spacing",
)
@click.option(
    "-latitude_grid_spacing",
    default=30,
    type=click.FLOAT,
    help="set y axis grid spacing",
)
@click.option(
    "-override_plot_properties",
    default=None,
    help='Override the following plot proporties: "cbar_shrink", "cbar_pad", "cbar_label_pad", "figure_width": width, "figure_size_ratio": ratio.',
)
@click.option(
    "-xtick_label_color",
    default="black",
    type=click.STRING,
    help="color of longitude ticks",
)
@click.option(
    "-ytick_label_color",
    default="black",
    type=click.STRING,
    help="color of latitude ticks",
)
@click.option(
    "-graticule_color",
    default=None,
    type=click.STRING,
    help="color of latitude ticks",
)
@click.option(
    "-fontsize",
    default=None,
    help="Dictionary with fontsizes: 'xlabel', 'ylabel', 'title', 'xtick_label', 'ytick_label', 'cbar_label', 'cbar_tick_label'.",
)
@click.option(
    "-phi_convention",
    default="counterclockwise",
    type=click.STRING,
    help='convention on x-axis (phi), "counterclockwise" (default), "clockwise", "symmetrical" (phi as it is truly given) if "flip" is "geo", "phi_convention" should be set to "clockwise".',
)
@click.option(
    "-custom_xtick_labels",
    default=None,
    help="override x-axis tick labels",
)
@click.option(
    "-custom_ytick_labels",
    default=None,
    help="override y-axis tick labels",
)
@click.option(
    "-white_background",
    is_flag=True,
    help="Sets the background to be white. (Transparent by default [recommended])",
)
@click.option(
    "-png",
    is_flag=True,
    help="Saves output as .png ().pdf by default)",
)
@click.option(
    "-outdir",
    default=None,
    type=click.Path(exists=True),
    help="Output directory for plot",
)
@click.option(
    "-outname",
    default=None,
    type=click.STRING,
    help="Output filename, overwrites autonaming",
)
@click.option(
    "-show",
    is_flag=True,
    help="Displays the figure in addition to saving",
)
@click.option(
    "-gif",
    is_flag=True,
    help="Convert inputs to gif",
)
@click.option(
    "-fps",
    default=3,
    type=click.INT,
    help="Gif speed, 300 by default.",
)
@click.option(
    "-dpi",
    default=300,
    type=click.INT,
    help="DPI of output file. 300 by default.",
)
def plot(
    input,
    sig,
    comp,
    freq,
    ticks,
    min,
    max,
    rng,
    nocbar,
    unit,
    fwhm,
    nside,
    sample,
    mask,
    maskfill,
    cmap,
    norm,
    remove_dip,
    remove_mono,
    title,
    right_label,
    left_label,
    width,
    xsize,
    darkmode,
    interactive,
    rot,
    coord,
    nest,
    flip,
    graticule,
    graticule_labels,
    return_only_data,
    projection_type,
    cb_orientation,
    xlabel,
    ylabel,
    longitude_grid_spacing,
    latitude_grid_spacing,
    override_plot_properties,
    xtick_label_color,
    ytick_label_color,
    graticule_color,
    fontsize,
    phi_convention,
    custom_xtick_labels,
    custom_ytick_labels,
    white_background,
    png,
    outdir,
    outname,
    show,
    gif,
    fps,
    dpi,
):
    """
    This function invokes the plot function from the command line
    input: filename of fits or h5 file, optional plotting paramameters
    """
    imgs = []  # Container for images used with gif
    for i, filename in enumerate(input):
        cbar = False if nocbar else True
        freq_ = None if freq is None else freq * u.GHz
        fwhm_ = None if fwhm is None else fwhm * u.arcmin
        comps = comp

        # Super hacky gif fix because "hold" is not implemented in newvisufunc
        if gif and i > 0:
            return_only_data = True

        # Actually plot the stuff
        img, params = cplot(
            filename,
            sig=sig,
            comp=comps,
            freq=freq_,
            ticks=ticks,
            min=min,
            max=max,
            rng=rng,
            cbar=cbar,
            unit=unit,
            fwhm=fwhm_,
            nside=nside,
            sample=sample,
            mask=mask,
            maskfill=maskfill,
            cmap=cmap,
            norm=norm,
            remove_dip=remove_dip,
            remove_mono=remove_mono,
            title=title,
            right_label=right_label,
            left_label=left_label,
            width=width,
            xsize=xsize,
            darkmode=darkmode,
            interactive=interactive,
            rot=rot,
            coord=coord,
            nest=nest,
            flip=flip,
            graticule=graticule,
            graticule_labels=graticule_labels,
            return_only_data=return_only_data,
            projection_type=projection_type,
            cb_orientation=cb_orientation,
            xlabel=xlabel,
            ylabel=ylabel,
            longitude_grid_spacing=longitude_grid_spacing,
            latitude_grid_spacing=latitude_grid_spacing,
            override_plot_properties=override_plot_properties,
            xtick_label_color=xtick_label_color,
            ytick_label_color=ytick_label_color,
            graticule_color=graticule_color,
            fontsize=fontsize,
            phi_convention=phi_convention,
            custom_xtick_labels=custom_xtick_labels,
            custom_ytick_labels=custom_ytick_labels,
        )

        # Super hacky gif fix because "hold" is not implemented in newvisufunc
        if gif:
            if i > 0:
                longitude, latitude, grid_map = img
                img = plt.pcolormesh(
                    longitude,
                    latitude,
                    grid_map,
                    vmin=params["ticks"][0],
                    vmax=params["ticks"][-1],
                    rasterized=True,
                    cmap=load_cmap(params["cmap"]),
                    shading="auto",
                )

            """
            # Show title for gifs
            sample = re.search('_k([0-9]*)', filename)
            if sample:
                sample = sample.group(1).lstrip("0")
            else:
                sample = filename
            """
        if show:
            plt.show()

        """
        Filename formatting and  outputting
        """
        *path, fn = os.path.split(filename)
        fn, ftype = os.path.splitext(fn)
        if ftype == ".h5":
            fn += "_" + comp

        fn = fn.replace("_IQU_", "_").replace("_I_", "_")
        if freq_ is not None:
            fn = fn.replace(comp, f"{comp}-{int(freq_.value)}GHz")
        if outdir:
            path = outdir

        tags = []
        tags.append(f"{str(int(fwhm_.value))}arcmin") if float(
            fwhm_.value
        ) > 0 else None
        tags.append("cb") if cbar else None
        tags.append("masked") if mask else None
        tags.append("nodip") if remove_dip else None
        tags.append("dark") if darkmode else None
        if params["cmap"] is None:
            params["cmap"] = "planck"
        tags.append(f'c-{params["cmap"]}')

        nside_tag = "_n" + str(int(params["nside"]))
        if nside_tag in fn:
            fn = fn.replace(nside_tag, "")

        if params["width"] is None:
            width = 8.5

        if isinstance(sig, int):
            sig = ["I", "Q", "U"][sig]
        fn = fn + f"_{sig}_w{str(int(width))}" + nside_tag

        for i in tags:
            fn += f"_{i}"

        filetype = "png" if png else "pdf"
        if outname:
            filetype = "png" if outname.endswith(".png") else "pdf"

        fn += f".{filetype}"
        if outname:
            fn = outname

        if outdir:
            path = [outdir]

        if gif:
            # Append images for animation
            imgs.append([img])
        else:
            tp = False if white_background else True
            filetype = "png" if png else "pdf"
            path = "." if path[0] == "" else path[0]
            print(f"[bold green]Outputting {fn}[/bold green] to {path}")
            plt.savefig(
                f"{path}/{fn}",
                bbox_inches="tight",
                pad_inches=0.02,
                transparent=tp,
                format=filetype,
                dpi=dpi,
            )

    if gif:
        import matplotlib.animation as animation

        ani = animation.ArtistAnimation(plt.gcf(), imgs, blit=True)
        fn = fn.replace(filetype, "gif")
        print(f"[bold green]Outputting {fn}[/bold green]")
        ani.save(fn, dpi=dpi, writer=animation.PillowWriter(fps=fps))


@commands_plotting.command()
@click.argument("input", type=click.STRING)
@click.option("-lon", default=0, type=click.INT)
@click.option("-lat", default=0, type=click.INT)
@click.option(
    "-comp",
    default=None,
    type=click.STRING,
    help="Component name for autosetting plottingparameters",
)
@click.option(
    "-sig",
    default=0,
    help="Which sky signal to plot",
)
@click.option(
    "-freq",
    default=None,
    type=click.FLOAT,
    help="Frequency in GHz needed for scaling maps when using a model object input",
)
@click.option(
    "-nside",
    default=None,
    type=click.INT,
    help="nside for optional ud_grade.",
)
@click.option("-size", default=20, type=click.INT)
@click.option(
    "-sample",
    default=-1,
    type=click.INT,
    help="Select sample when inputing .h5 file",
)
@click.option(
    "-min",
    "vmin",
    default=None,
    type=click.FLOAT,
    help="Min value of colorbar, overrides autodetector.",
)
@click.option(
    "-max",
    "vmax",
    default=None,
    type=click.FLOAT,
    help="Max value of colorbar, overrides autodetector.",
)
@click.option(
    "-ticks",
    default=None,
    help="Min and max value for data. If None, uses 97.5th percentile.",
)
@click.option("-range", "rng", default=None, type=click.FLOAT, help="Color bar range")
@click.option(
    "-left_label",
    default=None,
    type=click.STRING,
    help="Set title, has LaTeX functionality. Ex. $\mu$",
)
@click.option(
    "-right_label",
    default=None,
    type=click.STRING,
    help="Set title, has LaTeX functionality. Ex. $\mu$",
)
@click.option(
    "-unit",
    default=None,
    type=click.STRING,
    help="Set unit (Under color bar), has LaTeX functionality. Ex. $\mu$",
)
@click.option(
    "-cmap",
    default=None,
    help="Choose different color map (string), such as Jet or planck",
)
@click.option(
    "-graticule",
    is_flag=True,
    help="Add graticule",
)
@click.option(
    "-norm",
    default=None,
    help='"Linear" is normal, "log" Normalizesusingsemi-logscale',
)
@click.option(
    "-nocbar",
    is_flag=True,
    help="remove colorbar",
)
@click.option(
    "-cbar_pad",
    default=0.04,
    type=click.FLOAT,
    help="Colorbar space",
)
@click.option(
    "-cbar_shrink",
    default=0.7,
    type=click.FLOAT,
    help="Colorbar size",
)
@click.option(
    "-fwhm",
    default=0.0,
    type=click.FLOAT,
    help="FWHM of smoothing, in arcmin.",
)
@click.option(
    "-remove_dip",
    default=None,
    type=click.STRING,
    help="Fits a dipole to the map and removes it. Specify mask or type auto.",
)
@click.option(
    "-remove_mono",
    default=None,
    type=click.STRING,
    help="Fits a monopole to the map and removes it.",
)
@click.option(
    "-fontsize",
    default=None,
    type=click.INT,
    help="Fontsize",
)
@click.option(
    "-figsize",
    default=None,
    help="Figure size",
)
@click.option(
    "-darkmode",
    is_flag=True,
    help="Turn on darkmode",
)
@click.option(
    "-fignum",
    default=None,
    type=click.INT,
    help="Figure number",
)
@click.option(
    "-subplot",
    default=None,
    help="Subplot (ex. (2,1,2))",
)
@click.option(
    "-hold",
    is_flag=True,
    help="Hold figure",
)
@click.option(
    "-reuse_axes",
    is_flag=True,
    help="Reuse axes",
)
@click.option(
    "-outdir",
    default=None,
    type=click.Path(exists=True),
    help="Output directory for plot",
)
@click.option(
    "-outname",
    default=None,
    type=click.STRING,
    help="Output filename, overwrites autonaming",
)
@click.option(
    "-show",
    is_flag=True,
    help="Displays the figure in addition to saving",
)
@click.option(
    "-png",
    is_flag=True,
    help="Saves output as .png ().pdf by default)",
)
@click.option(
    "-dpi",
    default=300,
    type=click.INT,
    help="DPI of output file. 300 by default.",
)
def gnom(
    input,
    lon,
    lat,
    comp,
    sig,
    freq,
    nside,
    size,
    sample,
    vmin,
    vmax,
    ticks,
    rng,
    left_label,
    right_label,
    unit,
    cmap,
    graticule,
    norm,
    nocbar,
    cbar_pad,
    cbar_shrink,
    fwhm,
    remove_dip,
    remove_mono,
    fontsize,
    figsize,
    darkmode,
    fignum,
    subplot,
    hold,
    reuse_axes,
    outdir,
    outname,
    show,
    png,
    dpi,
):
    """
    This function plots a fits file in gnomonic view.
    It is wrapper on the cosmoglobe "gnom" function.s
    """
    cbar = False if nocbar else True
    freq_ = None if freq is None else freq * u.GHz
    fwhm_ = None if fwhm is None else fwhm * u.arcmin
    figsize = (3, 4) if figsize is None else figsize
    img, params = cgnom(
        input,
        lon,
        lat,
        comp,
        sig,
        freq_,
        nside,
        size,
        sample,
        vmin,
        vmax,
        ticks,
        rng,
        left_label,
        right_label,
        unit,
        cmap,
        graticule,
        norm,
        cbar,
        cbar_pad,
        cbar_shrink,
        fwhm_,
        remove_dip,
        remove_mono,
        fontsize,
        figsize,
        darkmode,
        fignum,
        subplot,
        hold,
        reuse_axes,
    )

    if show:
        plt.show()
    else:
        *path, fn = os.path.split(input)
        if outname:
            filetype = "png" if outname.endswith(".png") else "pdf"
        else:
            fn = (
                os.path.splitext(fn)[0]
                + f'_gnomonic_{["I","Q","U"][sig]}_{lon}lon{lat}lat_{size}x{size}deg_c-{params["cmap"]}'
            )
            filetype = "png" if png else "pdf"

        if outname:
            fn = outname
        if outdir:
            path = [outdir]
        path = "." if path[0] == "" else path[0]
        fn = os.path.splitext(fn)[0] + f".{filetype}"
        print(f"[bold green]Outputting {fn}[/bold green] to {path}")
        plt.savefig(
            f"{path}/{fn}", bbox_inches="tight", pad_inches=0.02, format=filetype, dpi=dpi
        )


@commands_plotting.command()
@click.argument(
    "input",
)
@click.option(
    "-dataset",
    default=None,
    type=click.STRING,
    help="HDF dataset to be plotted",
)
@click.option(
    "-sig",
    default=0,
    type=click.INT,
    help="Signal to be plotted 0 by default (0, 1, 2 is interprated as IQU)",
)
@click.option(
    "-labels",
    default=None,
    help="List of label values for each input",
)
@click.option(
    "-hideval",
    is_flag=True,
    help="Hide mean and standard deviation of samples",
)
@click.option(
    "-burnin",
    default=0,
    type=click.INT,
    help="Burn in",
)
@click.option(
    "-xlabel",
    default=None,
    type=click.STRING,
    help="set x axis label",
)
@click.option(
    "-ylabel",
    default=None,
    type=click.STRING,
    help="set y axis label",
)
@click.option(
    "-nbins",
    default=None,
    type=click.INT,
    help="Number of bins",
)
@click.option(
    "-cmap",
    default="tab10",
    type=click.STRING,
    help="Color map (tab10 by default)",
)
@click.option(
    "-figsize",
    default=None,
    help="figure size",
)
@click.option(
    "-darkmode",
    is_flag=True,
    help="Turn on darkmode",
)
@click.option(
    "-fignum",
    default=None,
    type=click.INT,
    help="Figure number",
)
@click.option(
    "-subplot",
    default=None,
    help="Subplot (ex. (2,1,2))",
)
@click.option(
    "-hold",
    is_flag=True,
    help="Hold figure",
)
@click.option(
    "-reuse_axes",
    is_flag=True,
    help="Reuse axes",
)
@click.option(
    "-outdir",
    default=None,
    type=click.Path(exists=True),
    help="Output directory for plot",
)
@click.option(
    "-outname",
    default=None,
    type=click.STRING,
    help="Output filename, overwrites autonaming",
)
@click.option(
    "-show",
    is_flag=True,
    help="Displays the figure in addition to saving",
)
@click.option(
    "-png",
    is_flag=True,
    help="Saves output as .png ().pdf by default)",
)
@click.option(
    "-dpi",
    default=300,
    type=click.INT,
    help="DPI of output file. 300 by default.",
)
def trace(
    input,
    dataset,
    sig,
    labels,
    hideval,
    burnin,
    xlabel,
    ylabel,
    nbins,
    cmap,
    figsize,
    darkmode,
    fignum,
    subplot,
    hold,
    reuse_axes,
    outdir,
    outname,
    show,
    png,
    dpi,
):
    """
    This function creates a trace plot off data in a HDF file.
    This is a click wrapper the same function in cosmoglobe.
    """
    labels = labels.split(" ")
    if figsize is None: figsize = (8, 3)
    showval=True if not hideval else False
    ctrace(
        input,
        dataset,
        sig,
        labels,
        showval,
        burnin,
        xlabel,
        ylabel,
        nbins,
        cmap,
        figsize,
        darkmode,
        fignum,
        subplot,
        hold,
        reuse_axes,
    )

    if show:
        plt.show()
    else:
        *path, fn = os.path.split(input)
        if outname:
            filetype = "png" if outname.endswith(".png") else "pdf"
        else:
            filetype = "png" if png else "pdf"

        if outname:
            fn = outname
        if outdir:
            path = [outdir]
        path = "." if path[0] == "" else path[0]
        fn = os.path.splitext(fn)[0] + f".{filetype}"
        print(f"[bold green]Outputting {fn}[/bold green] to {path}")
        plt.savefig(
            f"{path}/{fn}", bbox_inches="tight", pad_inches=0.02, format=filetype, dpi=dpi
        )



import healpy as hp
import matplotlib.pyplot as plt

from functools import partial
from .plottools import *


@u.quantity_input(freq=u.Hz, fwhm=(u.arcmin, u.rad, u.deg))
def gnom(
    input,
    lon=0,
    lat=0,
    comp=None,
    sig=0,
    freq=None,
    nside=None,
    size=20,
    sample=-1,
    vmin=None,
    vmax=None,
    ticks=None,
    rng=None,
    left_label=None,
    right_label=None,
    unit=None,
    cmap=None,
    graticule=False,
    norm=None,
    cbar=True,
    cbar_pad=0.04,
    cbar_shrink=0.7,
    fwhm=0.0 * u.arcmin,
    remove_dip=False,
    remove_mono=False,
    fontsize=None,
    figsize=(3, 4),
    darkmode=False,
    fignum=None,
    subplot=None,
    hold=False,
    reuse_axes=False,
):
    """
    Gnomonic view plotting.
    """
    if not fontsize:
        fontsize = {
            "xlabel": 11,
            "ylabel": 11,
            "xtick_label": 8,
            "ytick_label": 8,
            "title": 12,
            "cbar_label": 11,
            "cbar_tick_label": 9,
            "left_label": 11,
            "right_label": 11,
        }
    # Set plotting rcParams
    set_style(darkmode)
    xsize = 1000
    reso = size * 60 / xsize

    # Get data
    m, comp, freq, nside  = get_data(input, sig, comp, freq, fwhm, nside=nside, sample=sample)
    if remove_dip:
        m = hp.remove_dipole(m, gal_cut=30, copy=True, verbose=True)
    if remove_mono:
        m = hp.remove_monopole(m, gal_cut=30, copy=True, verbose=True)

    proj = hp.projector.GnomonicProj(
        rot=[lon, lat, 0.0], coord="G", xsize=xsize, ysize=xsize, reso=reso
    )
    reproj_im = proj.projmap(m, vec2pix_func=partial(hp.vec2pix, nside))

    # Pass all your arguments in, return parsed plotting parameters
    params = get_params(
        data=reproj_im, 
        comp=comp,
        sig=sig,
        right_label=right_label,
        left_label=left_label,
        unit=unit,
        ticks=ticks,
        min=vmin,
        max=vmax,
        rng=rng,
        norm=norm,
        cmap=cmap,
        freq_ref=freq,
        width=figsize[0],
        nside=nside
    )

    # Semi-log normalization
    if params["norm"] == "log":
        params["data"], params["ticks"] = apply_logscale(params["data"], params["ticks"], linthresh=1)

    cmap = load_cmap(params["cmap"])

    fig, ax = make_fig(
        figsize,
        fignum,
        hold,
        subplot,
        reuse_axes
    )
    image = plt.imshow(
        params["data"],
        origin="lower",
        interpolation="nearest",
        vmin=params["ticks"][0],
        vmax=params["ticks"][-1],
        cmap=cmap
    )

    plt.xticks([])
    plt.yticks([])
    plt.text(
        0.95,
        0.95,
        params["right_label"],
        color="black",
        va="top",
        ha="right",
        transform=ax.transAxes,
        fontsize=fontsize["title"]
    )
    plt.text(
        0.05,
        0.95,
        params["left_label"],
        color="black",
        va="top",
        transform=ax.transAxes,
        fontsize=fontsize["title"]
    )

    if cbar:
        apply_colorbar(
            plt.gcf(),
            plt.gca(),
            image,
            params["ticks"],
            params["ticklabels"],
            params["unit"],
            fontsize=fontsize,
            linthresh=1,
            norm=params["norm"],
            cbar_pad=cbar_pad,
            cbar_shrink=cbar_shrink
        )

    if graticule:
        hp.graticule()

    return image, params



from enum import Enum, auto
from typing import List, Iterator, Optional

import astropy.units as u
import healpy as hp
import numpy as np

from cosmoglobe.sky.base import SkyComponent, Diffuse, PointSource
from cosmoglobe.utils.utils import str_to_astropy_unit
from cosmoglobe.sky.exceptions import (
    ModelComponentNotFoundError,
    ModelComponentError,
)


class State(Enum):
    """State of a model component."""

    ENABLED = auto()
    DISABLED = auto()


class Model:
    r"""Sky model object representing the Cosmoglobe Sky Model.

    This class acts as a container for the various components making up
    the Cosmoglobe Sky Model, and provides methods to simulate the sky.
    The primary use case of this class is to call its ``__call__``
    method, which simulates the sky at a single frequency :math:`\nu`,
    or integrated over a bandpass :math:`\tau`.

    Attributes
    ----------
    ame : :class:`cosmoglobe.sky.components.AME`
        The AME sky component.
    cmb : :class:`cosmoglobe.sky.components.CMB`
        The CMB sky component.
    dust : :class:`cosmoglobe.sky.components.Dust`
        The dust sky component.
    ff : :class:`cosmoglobe.sky.components.FreeFree`
        The free-free sky component.
    radio : :class:`cosmoglobe.sky.components.Radio`
        The radio sky component.
    synch : :class:`cosmoglobe.sky.components.Synchrotron`
        The synchrotron sky component.
    nside : int
        Healpix resolution of the maps in sky model.

    Methods
    -------
    __call__
    disable
    enable

    See Also
    --------
    __call__ :
        Simulates the sky at a given frequency :math:`\nu` or over a
        bandpass :math:`\tau` given the Cosmoglobe Sky Model.

    Examples
    --------
    Inspecting the model:

    >>> model = skymodel(nside=256)
    >>> print(model)
    Model(
      nside: 256
      components(
        (ame): AME(nu_p)
        (cmb): CMB()
        (dust): Dust(beta, T)
        (ff): FreeFree(Te)
        (radio): Radio(specind)
        (synch): Synchrotron(beta)
      )
    )

    Simulating the full sky emission at some frequency, given a beam
    FWHM:

    >>> import astropy.units as u
    >>> emission = model(50*u.GHz, fwhm=30*u.arcmin)
    >>> print(emission)
    Smoothing point sources...
    Smoothing diffuse emission...
    [[ 2.25809786e+03  2.24380103e+03  2.25659060e+03 ... -2.34783682e+03
      -2.30464421e+03 -2.30387946e+03]
     [-1.64627550e+00  2.93583564e-01 -1.06788937e+00 ... -1.64354585e+01
       1.60621841e+01 -1.05506092e+01]
     [-4.15682825e+00  3.08881971e-01 -1.02012415e+00 ...  5.44745701e+00
      -4.71776995e+00  4.39850830e+00]] uK
    """

    def __init__(
        self,
        nside: Optional[int] = None,
        components: Optional[List[SkyComponent]] = None,
    ) -> None:
        """Initializing a sky model.

        Parameters
        ----------
        nside
            Healpix resolution of the maps in sky model. If None, nside
            will match the first component added to the model. Defaults to
            None.
        components
            A list of `SkyComponent` objects that constitutes the sky
            model (by default this is None and the components are
            iteratively added as they are read from a commander3 chain).
        """

        self._nside = nside
        self._components = {}

        if components is not None:
            for component in components:
                self._add_component_to_model(component)

    @property
    def nside(self) -> int:
        """Model nside."""

        return self._nside

    @property
    def components(self) -> List[SkyComponent]:
        """List of all enabled components."""

        return [
            comp[0] for comp in self._components.values() if comp[1] is State.ENABLED
        ]

    @u.quantity_input(
        freqs=u.Hz, bandpass=(u.Jy / u.sr, u.K, None), fwhm=(u.rad, u.deg, u.arcmin)
    )
    def __call__(
        self,
        freqs: u.Quantity,
        bandpass: Optional[u.Quantity] = None,
        fwhm: u.Quantity = 0.0 * u.rad,
        output_unit: u.UnitBase = u.uK,
    ) -> u.Quantity:
        r"""Simulates and returns the summed emission over all sky components. 

        The emission is computed for either a single frequency :math:`\nu` or 
        integrated over a  bandpass :math:`\tau`.

        Parameters
        ----------
        freqs
            A frequency, or a list of frequencies for which to simulate the
            sky emission.
        bandpass
            Bandpass profile corresponding to the frequencies in `freqs`. 
            If `bandpass` is None and `freqs` is a single frequency, a 
            delta peak is assumed. Defaults to None. 
        fwhm
            The full width half max parameter of the Gaussian. Defaults to
            0.0, which indicates no smoothing of output maps.
        output_unit
            The desired output units of the emission. Supported units are
            :math:`\mathrm{\mu K_{RJ}}` and :math:`\mathrm{MJ/sr}`. 
            Defaults to :math:`\mathrm{\mu K_{RJ}}`. 

        Returns
        -------
            The summed emission over all sky components.

        Notes
        -----
        This function computes the following expression (assuming that all
        components are enabled in the model):

        .. math::

            \boldsymbol{s}_{\mathrm{RJ}}(\nu) &=\boldsymbol{a}_{\mathrm{CMB}}
             \frac{x^{2} \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)^{2}} 
            \frac{\left(\mathrm{e}^{x_{0}}-1\right)^{2}}{x_{0}^{2} 
            \mathrm{e}^{x_{0}}}\\
            &+\boldsymbol{a}_{\mathrm{s}}\left(\frac{\nu}{\nu_{0, 
            \mathrm{~s}}}\right)^{\beta_{\mathrm{s}}}\\
            &+\boldsymbol{a}_{\mathrm{ff}} \frac{g_{\mathrm{ff}}
            \left(\nu ; T_{e}\right)}{g_{\mathrm{ff}}\left(\nu_{0, 
            \mathrm{ff}} ; T_{e}\right)}\left(\frac{\nu_{0, 
            \mathrm{ff}}}{\nu}\right)^{2}\\
            &+\boldsymbol{a}_{\mathrm{sd}}\left(\frac{\nu_{0, 
            \mathrm{sd}}}{\nu}\right)^{2} \frac{s_{0}^{\mathrm{sd}}
            \left(\nu \cdot \frac{\nu_{p}}{30.0\; \mathrm{GHz}}\right)}
            {s_{0}^{\mathrm{sd}}\left(\nu_{0, \mathrm{sd}} \cdot 
            \frac{\nu_{p}}{30.0\; \mathrm{GHz}}\right)} \\
            &+\boldsymbol{a}_{\mathrm{d}}\left(\frac{\nu}{\nu_{0, 
            \mathrm{~d}}}\right)^{\beta_{\mathrm{d}}+1} 
            \frac{\mathrm{e}^{h \nu_{0, \mathrm{~d}} 
            / k T_{\mathrm{d}}}-1}{\mathrm{e}^{\mathrm{h} \nu 
            / k T_{\mathrm{d}}}-1}+\\&+\sum_{j=1}^{N_{\mathrm{src}}} 
            \boldsymbol{a}_{\mathrm{src}}^{j}\left(\frac{\nu}{\nu_{0, 
            \mathrm{src}}}\right)^{\alpha_{j, \mathrm{src}}-2}.

        For more information on the current implementation of the 
        Cosmoglobe Sky Model, see `BeyondPlanck (2020), Section 3.6 
        <https://arxiv.org/pdf/2011.05609.pdf>`_.

        Examples
        --------
        Simulated full sky emission at :math:`50\; \mathrm{GHz}`:

        >>> from cosmoglobe import skymodel
        >>> import astropy.units as u
        >>> model = skymodel(nside=256) 
        >>> model(50*u.GHz)[0]
        [ 2234.74893115  2291.99921295  2323.98779311 ... -2320.74732945
         -2271.54465982 -2292.22248419] uK

        Simulated full sky emission at :math:`500\; \mathrm{GHz}` 
        smoothed with a :math:`50\; '` Gaussian beam, outputed in units of 
        :math:`\mathrm{MJy} / \mathrm{sr}`:

        >>> model(500*u.GHz, fwhm=50*u.arcmin, output_unit='MJy/sr')[0]
        [ 0.267749    0.26991688  0.28053964 ... -0.15846278 -0.15269807
         -0.14408377] MJy / sr
        """

        diffuse_emission = np.zeros((3, hp.nside2npix(self._nside)))
        point_source_emission = np.zeros_like(diffuse_emission)

        # The output unit may be a string denoting for instance K_CMB, which
        # is critical information for the following routines. However, we
        # need to initialize the emission arrays with astropy units.
        _output_unit = str_to_astropy_unit(output_unit)
        diffuse_emission = u.Quantity(diffuse_emission, unit=_output_unit)
        point_source_emission = u.Quantity(point_source_emission, unit=_output_unit)

        for comp in self.components:
            if isinstance(comp, Diffuse):
                comp_emission = comp(freqs, bandpass, output_unit=output_unit)
                for idx, row in enumerate(comp_emission):
                    diffuse_emission[idx] += row

            elif isinstance(comp, PointSource):
                comp_emission = comp(
                    freqs, bandpass, fwhm=fwhm, output_unit=output_unit
                )
                for idx, row in enumerate(comp_emission):
                    point_source_emission[idx] += row

        if fwhm != 0.0 and diffuse_emission.value.any():
            # Instead of calling the diffuse components _smooth function
            # it is more efficient to sum all the emission first and then
            # only perform a singled smoothing operation.
            print("Smoothing diffuse emission...")
            diffuse_emission = u.Quantity(
                hp.smoothing(diffuse_emission, fwhm.to(u.rad).value),
                unit=diffuse_emission.unit,
            )

        return diffuse_emission + point_source_emission

    def disable(self, comp_label: str) -> None:
        """Disable a component in the model.

        Parameters
        ----------
        comp_label
            The component label.

        Raises
        ------
        KeyError
            If the component is not currently enabled in the model.
        """

        try:
            if self._components[comp_label][1] is State.ENABLED:
                self._components[comp_label][1] = State.DISABLED
            else:
                raise ModelComponentError(f"{comp_label!r} is already disabled")
        except KeyError:
            raise ModelComponentNotFoundError(
                f"{comp_label!r} is not present in the model"
            )

    def enable(self, comp_label: str) -> None:
        """Enable a disabled component.

        Parameters
        ----------
        comp_label
            The component label.

        Raises
        ------
        KeyError
            If the component is not currently disabled in the model.
        """

        try:
            if self._components[comp_label][1] is State.DISABLED:
                self._components[comp_label][1] = State.ENABLED
            else:
                raise ModelComponentError(f"{comp_label!r} is already enabled")
        except KeyError:
            raise ModelComponentNotFoundError(
                f"{comp_label!r} is not a component in the model"
            )

    def _add_component_to_model(self, component: SkyComponent) -> None:
        """Adds a new component to the model."""

        if not isinstance(component, SkyComponent):
            raise TypeError(f"component must be a subclass of {SkyComponent}")
        name = component.label
        if name is None:
            raise ValueError("Cannot add component without label")
        if name in self._components:
            raise ModelComponentError(
                f"component {name} is already a part of the model"
            )

        setattr(self, name, component)
        self._components[name] = [component, State.ENABLED]

        if self._nside is None:
            self._nside = hp.get_nside(component.amp)

        if hasattr(component, "_nside"):
            component._nside = self._nside

    def __iter__(self) -> Iterator:
        """Returns an iterator with active model components"""

        return iter(self.components)

    def __repr__(self) -> str:
        """Representation of the Model and all enabled components."""

        reprs = []
        for component in self.components:
            component_repr = repr(component) + "\n"
            reprs.append(f"({component.label}): {component_repr}")

        main_repr = f"Model("
        main_repr += f"\n  nside: {self._nside}"
        main_repr += "\n  components( "
        main_repr += "\n    " + "    ".join(reprs)
        main_repr += f"  )"
        main_repr += f"\n)"

        return main_repr



import astropy.units as u
import numpy as np

import cosmoglobe.utils.constants as const


@u.quantity_input(freq=u.Hz, T=u.K)
def blackbody_emission(freq: u.Quantity, T: u.Quantity) -> u.Quantity:
    """Returns the blackbody emission.

    Computes the emission emitted by a blackbody with with temperature
    T at a frequency freq in SI units [W / m^2 Hz sr].

    Parameters
    ----------
    freq
        Frequency [Hz].
    T
        Temperature of the blackbody [K].

    Returns
    -------
    emission
        Blackbody emission [W / m^2 Hz sr].
    """

    # Avoiding overflow and underflow.
    T = T.astype(np.float64)

    term1 = (2 * const.h * freq ** 3) / const.c ** 2
    term2 = np.expm1((const.h * freq) / (const.k_B * T))

    emission = term1 / term2 / u.sr

    return emission.to("W / m^2 Hz sr")


@u.quantity_input(freq=u.Hz, T_e=u.K)
def gaunt_factor(freq: u.Quantity, T_e: u.Quantity) -> u.Quantity:
    """Returns the Gaunt factor.

    Computes the gaunt factor for a given frequency and electron
    temperaturein SI units.

    Parameters
    ----------
    freq
        Frequency [Hz].
    T_e
        Electron temperature [K].

    Returns
    -------
    gaunt_factor
        Gaunt Factor.
    """

    # Avoiding overflow and underflow.
    T_e = T_e.astype(np.float64)
    T_e = (T_e.to(u.kK)).value / 10
    freq = (freq.to(u.GHz)).value

    gaunt_factor = np.log(
        np.exp(5.96 - (np.sqrt(3) / np.pi) * np.log(freq * T_e ** -1.5)) + np.e
    )

    return u.Quantity(gaunt_factor, unit=u.dimensionless_unscaled)


@u.quantity_input(freq=u.Hz, T=u.K)
def thermodynamical_to_brightness(
    freq: u.Quantity, T: u.Quantity = const.T_0
) -> u.Quantity:
    """Conversion factor between K_CMB and K_RJ.

    Returns the conversion factor between thermodynamical and brightness
    temperatures [K_CMB -> K_RJ].

    Parmaeters
    ----------
    freq
        Frequency [K].

    Returns:
    --------
        Conversion factor.
    """

    x = (const.h * freq) / (const.k_B * T)

    return ((x ** 2 * np.exp(x)) / (np.expm1(x) ** 2)).si


@u.quantity_input(freq=u.Hz, T=u.K)
def brightness_to_thermodynamical(
    freq: u.Quantity, T: u.Quantity = const.T_0
) -> u.Quantity:
    """Conversion factor between K_RJ and K_CMB.

    Returns the conversion factor between thermodynamical and brightness
    temperatures [K_RJ -> K_CMB].

    Parmaeters
    ----------
    freq
        Frequency [K].

    Returns:
    --------
        Conversion factor.
    """

    return 1 / thermodynamical_to_brightness(freq, T)



import click
from cosmoglobe.plot.click_plotting import commands_plotting

CONTEXT_SETTINGS = dict(max_content_width=100)

@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass

cli = click.CommandCollection(sources=[commands_plotting,], context_settings=CONTEXT_SETTINGS)

if __name__ == '__main__':
    cli()
    


import inspect
from typing import List, Union, Optional, Literal

import astropy.units as u
import healpy as hp
from tqdm import tqdm

from cosmoglobe.h5.chain import Chain, ChainVersion
from cosmoglobe.h5.chain_context import chain_context
from cosmoglobe.h5.exceptions import (
    ChainComponentNotFoundError,
    ChainKeyError,
    ChainFormatError,
)
from cosmoglobe.sky.model import Model
from cosmoglobe.sky.base import SkyComponent
from cosmoglobe.sky import COSMOGLOBE_COMPS


DEFAULT_SAMPLE = -1


def model_from_chain(
    chain: Union[str, Chain],
    components: Optional[List[str]] = None,
    nside: Optional[int] = None,
    samples: Optional[Union[range, int, Literal["all"]]] = DEFAULT_SAMPLE,
    burn_in: Optional[int] = None,
) -> Model:
    """Initialize and return a cosmoglobe sky model from a chainfile.

    Parameters
    ----------
    chain
        Path to a Cosmoglobe chainfile or a Chain object.
    components
        List of components to include in the model.
    nside
        Model HEALPIX map resolution parameter.
    samples
        The sample number for which to extract the model. If the input
        is 'all', then the model will an average of all samples in the chain.
        Defaults to the last sample in the chain.
    burn_in
        Burn in sample for which all previous samples are disregarded.

    Returns
    -------
    model
        Initialized sky mode object.
    """

    if not isinstance(chain, Chain):
        chain = Chain(chain, burn_in)

    if isinstance(samples, str):
        if samples.lower() == "all":
            samples = None
        else:
            raise ValueError("samples must be either 'all', an int, or a range")

    if chain.version is ChainVersion.OLD:
        raise ChainFormatError(
            "cannot initialize a sky model from a chain without a " "parameter group"
        )

    print(f"Initializing model from {chain.path.name}")
    if components is None:
        components = chain.components
    elif any(component not in chain.components for component in components):
        raise ChainComponentNotFoundError(f"component was not found in chain")

    initialized_components: List[SkyComponent] = []
    with tqdm(total=len(components), ncols=75) as progress_bar:
        padding = len(max(chain.components, key=len))
        for component in components:
            progress_bar.set_description(f"{component:<{padding}}")
            initialized_components.append(
                _comp_from_chain(chain, component, nside, samples)
            )
            progress_bar.update()

    return Model(nside=nside, components=initialized_components)


def _comp_from_chain(
    chain: Chain,
    component: str,
    nside: Optional[int] = None,
    samples: Optional[Union[range, int]] = None,
) -> SkyComponent:
    """Initialize and return a sky component from a chainfile.

    Parameters
    ----------
    component
        Name of the component in the chain.
    chain
        Chain object.
    nside
        Model HEALPIX map resolution parameter.

    Returns
    -------
        Initialized sky component object.
    """
    try:
        comp_class = COSMOGLOBE_COMPS[component]
    except KeyError:
        raise ChainComponentNotFoundError(
            f"{component=!r} is not part in the Cosmoglobe Sky Model"
        )

    signature = inspect.signature(comp_class)
    class_args = list(signature.parameters.keys())

    args = {}

    # Contexts are operations that needs to be done to the data in the
    # chain before it can be used in the sky model.
    mappings = chain_context.get_mappings(component)
    units = chain_context.get_units(component)

    for arg in class_args:
        chain_arg = mappings.get(arg, arg)
        chain_params = chain.parameters[component]

        if chain_arg in chain_params:
            value = chain_params[chain_arg]
        else:
            try:
                value = chain.mean(f"{component}/{chain_arg}_alm", samples=samples)
                is_alm = True
            except ChainKeyError:
                try:
                    value = chain.mean(f"{component}/{chain_arg}", samples=samples)
                except ChainKeyError:
                    value = chain.mean(f"{component}/{chain_arg}_map", samples=samples)
                is_alm = False

            if is_alm:
                pol = True if arg == "amp" and value.shape[0] == 3 else False
                value = hp.alm2map(
                    value,
                    nside=nside if nside is not None else chain_params["nside"],
                    fwhm=(chain_params["fwhm"] * u.arcmin).to("rad").value,
                    pol=pol,
                    pixwin=True,
                )

        args[arg] = u.Quantity(value, unit=units[arg] if arg in units else None)

    contexts = chain_context.get_context(component)
    for context in contexts:
        args.update(context(args))

    return comp_class(**args)



from typing import Dict, Iterable

import astropy.units as u
import numpy as np

import cosmoglobe.utils.functions as F


def extract_scalars(spectral_parameters: Dict[str, u.Quantity]):
    """Extracts scalars from a spectral_parameter dict.

    If a map is constant in all dims, then it is converted to a (ndim,)
    array.
    """

    scalars = {}
    for key, value in spectral_parameters.items():
        if not isinstance(value, u.Quantity):
            value = u.Quantity(value, unit=u.dimensionless_unscaled)
        if value.size > 1:
            if value.ndim > 1:
                uniques = [np.unique(col) for col in value.value]
                all_is_unique = all([len(col) == 1 for col in uniques])
            else:
                uniques = np.unique(value.value)
                all_is_unique = len(uniques) == 1
            if all_is_unique:
                scalars[key] = uniques * value.unit
        else:
            scalars[key] = value

    return scalars


def set_spectral_units(maps: Iterable[u.Quantity]) -> u.Quantity:
    """Sets the spectral value for a specific spectral parameter name.

    Parameters
    ----------
    maps
        List of maps.

    Returns
    -------
    maps
        List of maps with new units.
    """

    # TODO: Figure out how to correctly detect unit of spectral map in chain.
    units = dict(
        T=u.K,
        Te=u.K,
        nu_p=u.GHz,
    )
    for map_ in maps:
        if map_ in units:
            maps[map_] *= units[map_]
        else:
            maps[map_] *= u.dimensionless_unscaled

    return maps


def str_to_astropy_unit(unit: str) -> u.UnitBase:
    """Converts a K_RJ or a K_CMB string to `u.K` unit.

    This function preserves the prefix.

    Parameters
    ----------
    unit
        String to convert to a astropy unit.

    Returns
    -------
    output_unit
        Astropy unit.
    """

    try:
        output_unit = u.Unit(unit)
    except ValueError:
        if unit.lower().endswith("k_rj"):
            output_unit = u.Unit(unit[:-3])

        elif unit.lower().endswith("k_cmb"):
            output_unit = u.Unit(unit[:-4])

    return output_unit


def to_unit(emission: u.Quantity, freqs: u.Quantity, unit: u.UnitBase) -> u.Quantity:
    """Converts the unit of the emission.

    Parameters
    ----------
    emission
        Emission whos unit we want to convert.
    freqs
        Frequency or a list of frequencies for which the emission was
        obtained.
    unit
        Unit to convert to.

    Returns
    -------
    emission
        Emission with new units.
    """

    try:
        unit = u.Unit(unit)
        emission = emission.to(unit, equivalencies=u.brightness_temperature(freqs))

    except ValueError:
        if unit.lower().endswith("k_rj"):
            unit = u.Unit(unit[:-3])
        elif unit.lower().endswith("k_cmb"):
            unit = u.Unit(unit[:-4])
            emission *= F.brightness_to_thermodynamical(freqs)
        emission.to(unit)

    return emission


def gaussian_beam_2D(r: np.ndarray, sigma: float) -> np.ndarray:
    """Returns the Gaussian beam in 2D in polar coordinates.

    Parameters
    ----------
    r
        Angular distance.
    sigma
        The sigma of the Gaussian (beam radius).

    Returns
    -------
        Gaussian beam.
    """

    return r * np.exp(-(r ** 2) / (2 * sigma ** 2)) / (sigma * np.sqrt(2 * np.pi))



import pathlib
from enum import Enum
from typing import Optional

from astropy.utils.data import download_file

# from cosmoglobe.chain.h5 import model_from_h5

data_url = pathlib.Path("http://cosmoglobe.uio.no/releases/")


class Release(Enum):
    """Class that enumerates Cosmoglobe data releases."""

    BP10 = LATEST = "BP10"


def sky_model(nside: int, release: Optional[str] = None, cache: bool = True) -> "Model":
    """Initialize the Cosmoglobe Sky Model from a official release.

    By default, the model is initialized from the latest stable cosmoglobe data
    release. The data required to initialize the model is downloaded and cached
    by default every time a unique nside is selected.

    For a list of available releases see the get_releases function.

    Parameters
    ----------
    nside : int
        Healpix resolution parameter to initialize the model at.
    release : int, str
        Data release version. By default, the latest stable release is selected,
        which corresponds to -1.
        Default: -1
    cache : bool
        Wether to cache the data after the download. Default: True.
    """

    filename = f"model_{nside}.h5"
    path = "/Users/metinsan/Documents/doktor/models/test1/"

    if release is None:
        release = Release.LATEST
    else:
        try:
            release = Release[release.upper()]
        except KeyError:
            raise ValueError(
                f"Release {release} does not exist. Available releases are: "
                f"{[name for name, _ in Release.__members__.items() if name != 'LATEST']}"
            )



if __name__ == "__main__":
    sky_model(64, "latesst")



import warnings
import os

from rich import print
import healpy as hp
import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt

from .plottools import *

# Fix for macos openMP duplicate bug
os.environ["KMP_DUPLICATE_LIB_OK"] = "True"

@u.quantity_input(freq=u.Hz, fwhm=(u.arcmin, u.rad, u.deg))
def plot(
    input,
    *,
    sig=0,
    comp=None,
    freq=None,
    ticks=None,
    min=None,
    max=None,
    rng=None,
    cbar=True,
    unit=None,
    fwhm=0.0 * u.arcmin,
    nside=None,
    sample=-1,
    mask=None,
    maskfill=None,
    cmap=None,
    norm=None,
    remove_dip=False,
    remove_mono=False,
    title=None,
    right_label=None,
    left_label=None,
    width=None,
    xsize=1000,
    darkmode=False,
    interactive=False,
    rot=None,
    coord=None,
    nest=False,
    flip="astro",
    graticule=False,
    graticule_labels=False,
    return_only_data=False,
    projection_type="mollweide",
    cb_orientation="horizontal",
    xlabel=None,
    ylabel=None,
    longitude_grid_spacing=60,
    latitude_grid_spacing=30,
    override_plot_properties=None,
    xtick_label_color="black",
    ytick_label_color="black",
    graticule_color=None,
    fontsize=None,
    phi_convention="counterclockwise",
    custom_xtick_labels=None,
    custom_ytick_labels=None,
):
    """
    General plotting function for healpix maps.
    This function is a wrapper on healpys projview function with some added features.
    Features added in addition to existing projview features include:
        - direct plotting from model object, fits file or numpy array.
        - simple map operations such as smoothing, ud_grading, masking
        - removing dipole or monopoles
        - Predefined plotting parameters for specific components for nice formatting

    Parameters
    ----------
    input : ndarray, fits file path or cosmoglobe model object
        Map data input given as numpy array either 1d or index given by 'sig'.
        Also supports fits-file path string or cosmoglobe model.
        If cosmoglobe object is passed such as 'model', specify comp or freq.
    sig : str or int, optional
        Specify which signal to plot if ndim>1.
        default = None
    comp : string, optional
        Component label for automatic identification of plotting
        parameters based on information from autoparams.json
        default = None
    freq : astropy GHz, optional
        frequency in GHz needed for scaling maps when using a model object input
        default = None
    ticks : list or str, optional
        Min and max value for data. If None, uses 97.5th percentile.
        default = None
    min : float, optional
      The minimum range value. If specified, overwrites autodetector.
      default = None
    max : float, optional
      The maximum range value. If specified, overwrites autodetector.
      default = None
    rng : float, optional
      Sets this value as min and max value. If specified, overwrites autodetector.
      default = None
    cbar : bool, optional
        Toggles the colorbar
        cbar = True
    fwhm : astropy arcmin/rad/deg, optional
        Optional map smoothing. FWHM of gaussian smoothing in arcmin.
        default = 0.0
    mask : str path or np.ndarray, optional
        Apply a mask file to data
        default = None
    cmap : str, optional
        Colormap (ex. sunburst, planck, jet). Both matplotliib and cmasher
        available as of now. Also supports qualitative plotly map, [ex.
        q-Plotly-4 (q for qualitative 4 for max color)] Sets planck as default.
        default = None
    norm : str, optional
        if norm=='linear':
            normal
        if norm=='log':
            Normalizes data using a semi-logscale linear between -1 and 1.
            Autodetector uses this sometimes, you will be warned.
        default = None
    remove_dip : bool, optional
        If mdmask is specified, fits and removes a dipole.
        default = True
    remove_mono : bool, optional
        If mdmask is specified, fits and removes a monopole.
        default = True
    unit : str, optional
        Unit label for colorbar
        default = None
    title : str, optional
        Sets the full figure title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default = None
    right_label : str, optional
        Sets the upper right title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default = None
    left_label : str, optional
        Sets the upper left title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default = None
    darkmode : bool, optional
        Plots all outlines in white for dark backgrounds, and adds 'dark' in
        filename.
        default = False
    rot : scalar or sequence, optional
      Describe the rotation to apply.
      In the form (lon, lat, psi) (unit: degrees) : the point at
      longitude *lon* and latitude *lat* will be at the center. An additional rotation
      of angle *psi* around this direction is applied.
    coord : sequence of character, optional
      Either one of 'G', 'E' or 'C' to describe the coordinate
      system of the map, or a sequence of 2 of these to rotate
      the map from the first to the second coordinate system.
    nest : bool, optional
      If True, ordering scheme is NESTED. Default: False (RING)
    flip : {'astro', 'geo'}, optional
      Defines the convention of projection : 'astro' (default, east towards left, west towards right)
      or 'geo' (east towards roght, west towards left)
      It creates the `healpy_flip` attribute on the Axes to save the convention in the figure.
    graticule : bool
      add graticule
    graticule_labels : bool
      longitude and latitude labels
    return_only_data : bool
      Return figure
    projection_type :  {'aitoff', 'hammer', 'lambert', 'mollweide', 'cart', '3d', 'polar'}
      type of the plot
    cb_orientation : {'horizontal', 'vertical'}
      color bar orientation
    xlabel : str
      set x axis label
    ylabel : str
      set y axis label
    longitude_grid_spacing : float
      set x axis grid spacing
    latitude_grid_spacing : float
      set y axis grid spacing
    override_plot_properties : dict
      Override the following plot proporties: 'cbar_shrink', 'cbar_pad', 'cbar_label_pad', 'figure_width': width, 'figure_size_ratio': ratio.
    lcolor : str
      change the color of the longitude tick labels, some color maps make it hard to read black tick labels
    fontsize:  dict
        Override fontsize of labels: 'xlabel', 'ylabel', 'title', 'xtick_label', 'ytick_label', 'cbar_label', 'cbar_tick_label'.
        default = None
    phi_convention : string
        convention on x-axis (phi), 'counterclockwise' (default), 'clockwise', 'symmetrical' (phi as it is truly given)
        if `flip` is 'geo', `phi_convention` should be set to 'clockwise'.
    custom_xtick_labels : list
        override x-axis tick labels
    custom_ytick_labels : list
        override y-axis tick labels
    """
    
    # Pick sizes from size dictionary for page width plots
    if width is None:
        override_plot_properties = None
    else:
        try:
            width = float(width)
        except ValueError:
            if isinstance(width, str):
                width = FIGURE_WIDTHS[width]
        ratio = 0.63 if cbar else 0.5
        xsize = int((1000 / 8.5) * width)
        override_plot_properties = {
            "figure_width": width,
            "figure_size_ratio": ratio,
            "cbar_pad": 0.04,
            "cbar_shrink": 0.3,
        }

    if not fontsize:
        fontsize = DEFAULT_FONTSIZES
    set_style(darkmode)

    # Translate sig to correct format
    if isinstance(sig, str):
        sig = STOKES.index(sig)


    # Get data
    m, comp, freq, nside = get_data(input, sig, comp, freq, fwhm, nside=nside, sample=sample)

    # Mask map
    if mask is not None:
        if isinstance(mask, str):
            mask = hp.read_map(mask)

        m = hp.ma(m)
        if hp.get_nside(mask) != nside:
            print(
                "[magenta]Input mask nside is different, ud_grading to output nside.[/magenta]"
            )
            mask = hp.ud_grade(mask, nside)
        m.mask = np.logical_not(mask)

    # Remove mono/dipole
    if remove_dip:
        m = hp.remove_dipole(
            m,
            gal_cut=30,
            copy=True,
        )
    if remove_mono:
        m = hp.remove_monopole(
            m,
            gal_cut=30,
            copy=True,
        )

    
    # Pass all your arguments in, return parsed plotting parameters
    params = get_params(
        data=m,
        comp=comp,
        sig=sig,
        right_label=right_label,
        left_label=left_label,
        unit=unit,
        ticks=ticks,
        min=min,
        max=max,
        rng=rng,
        norm=norm,
        cmap=cmap,
        freq_ref=freq,
        width=width,
        nside=nside,
    )

    # Semi-log normalization
    if params["norm"] == "log":
        params["data"], params["ticks"] = apply_logscale(
            params["data"], params["ticks"], linthresh=1
        )

    # Colormap
    cmap = load_cmap(params["cmap"])
    if maskfill:
        cmap.set_bad(maskfill)

    
    if interactive: # Plot using mollview if interactive mode
        if params["norm"] == "log":
            params["norm"] = None
        hp.mollview(
            params["data"],
            min=params["ticks"][0],
            max=params["ticks"][-1],
            cbar=False,
            cmap=cmap,
            unit=params["unit"],
            title=title,
            norm=params["norm"],
            # unedited params
            rot=rot,
            coord=coord,
            nest=nest,
            flip=flip,
            return_projected_map=True,
        )
        ret=plt.gca().get_images()[0]

    else: # Using fancy projview

        warnings.filterwarnings("ignore")  # Healpy complains too much
        # Plot figure
        ret = hp.newvisufunc.projview(
            params["data"],
            min=params["ticks"][0],
            max=params["ticks"][-1],
            cbar=False,
            cmap=cmap,
            xsize=xsize,
            # unedited params
            title=title,
            rot=rot,
            coord=coord,
            nest=nest,
            flip=flip,
            graticule=graticule,
            graticule_labels=graticule_labels,
            return_only_data=return_only_data,
            projection_type=projection_type,
            cb_orientation=cb_orientation,
            xlabel=xlabel,
            ylabel=ylabel,
            longitude_grid_spacing=longitude_grid_spacing,
            latitude_grid_spacing=latitude_grid_spacing,
            override_plot_properties=override_plot_properties,
            xtick_label_color=xtick_label_color,
            ytick_label_color=ytick_label_color,
            graticule_color=graticule_color,
            fontsize=fontsize,
            phi_convention=phi_convention,
            custom_xtick_labels=custom_xtick_labels,
            custom_ytick_labels=custom_ytick_labels,
        )
        # Remove color bar because of healpy bug
        plt.gca().collections[-1].colorbar.remove()
        
    
    if not return_only_data:

        if override_plot_properties is not None:
            cbar_pad = override_plot_properties["cbar_pad"]
            cbar_shrink = override_plot_properties["cbar_shrink"]
        else: 
            cbar_pad = 0.04
            cbar_shrink = 0.3

        # Add pretty color bar
        if cbar:
            apply_colorbar(
                plt.gcf(),
                plt.gca(),
                ret,
                params["ticks"],
                params["ticklabels"],
                params["unit"],
                fontsize=fontsize,
                linthresh=1,
                norm=params["norm"],
                cbar_pad=cbar_pad,
                cbar_shrink=cbar_shrink

            )

    #### Right Title ####
    plt.text(
        0.925,
        0.925,
        params["right_label"],
        ha="center",
        va="center",
        fontsize=fontsize["right_label"],
        transform=plt.gca().transAxes,
    )
    #### Left Title (stokes parameter label by default) ####
    plt.text(
        0.075,
        0.925,
        params["left_label"],
        ha="center",
        va="center",
        fontsize=fontsize["left_label"],
        transform=plt.gca().transAxes,
    )


    return ret, params




from cosmoglobe.plot.plottools import standalone_colorbar
from cosmoglobe.plot import plot, gnom, trace
from cosmoglobe.h5.model import model_from_chain
from cosmoglobe.h5.chain import Chain
from astropy import constants as const
import numpy as np 
import healpy as hp 
import astropy.units as u
import matplotlib.pyplot as plt
import os

paperfigs='/Users/svalheim/work/BP/papers/09_leakage/figs/'
path='/Users/svalheim/work/cosmoglobe-workdir/'
chain='bla.h5'
dust='dust_c0001_k000200.fits'

def transform(m):
    y = m/100
    y[y > 1] = np.log10(y[y>1]) + 1
    return y

def f(nu):
    return 1/(const.c.value**2 / 2. / const.k_B.value / (nu * 1e9) **2 * 1e-26 * 1e6)

def f2(nu):
    return (nu/22)**-3

if True:
    model = model_from_chain(path+chain, components=["dust"], nside=16)
    plot(model)
    plt.show()

if True:
    plot(path+chain, comp="dust", nside=16, )
    plt.show()
    #plot(path+dust, comp="dust", nside=16, )

if True:
    plot(path+dust, comp="dust",  sig="U", interactive=True) 
    plt.show()

if True:
    plot(path+dust, sig="Q", ticks=[0, np.pi, None], norm="log", unit=u.uK, fwhm=14*u.arcmin, nside=512, cmap="chroma", left_label="Pee", right_label="Poo", title="Cool figure", width=7, graticule=True, projection_type="hammer", mask=path+"mask_common_dx12_n0512_TQU.fits", maskfill="pink", graticule_color="white", xtick_label_color="white", ytick_label_color="white",)# darkmode=True)
    plt.show()

if True:
    plot(path+dust, interactive=True, sig="Q", ticks=[0, np.pi, None], norm="log", unit=u.uK, fwhm=14*u.arcmin, nside=512, cmap="chroma", left_label="Pee", right_label="Poo", title="Cool figure", width=7, graticule=True, projection_type="hammer", mask=path+"mask_common_dx12_n0512_TQU.fits", maskfill="pink", graticule_color="white", xtick_label_color="white", ytick_label_color="white",)# darkmode=True)
    plt.show()

if True:
    plot(path+dust, comp="cmb", sig="Q", ticks=[0, np.pi, None], norm="log", unit=u.uK, fwhm=14*u.arcmin, nside=512, cmap="chroma", left_label="Pee", right_label="Poo", title="Cool figure", width=7, graticule=True, projection_type="hammer", mask=path+"mask_common_dx12_n0512_TQU.fits", maskfill="pink", graticule_color="white", xtick_label_color="white", ytick_label_color="white",)# darkmode=True)
    plt.show()

if True:
    m = hp.read_map(path+dust, field=1)
    plot(m, comp="dust",  sig="U") 
    plt.show()

if True:
    gnom(path+dust, comp="dust", subplot=(2,2,1), cbar_pad=0.0, cbar_shrink=0.7)
    gnom(path+dust, comp="dust", subplot=(2,2,2), cbar_pad=0.0, cbar_shrink=0.8)
    gnom(path+dust, comp="dust", subplot=(2,2,3), cbar_pad=0.0, cbar_shrink=0.9)
    gnom(path+dust, comp="dust", subplot=(2,2,4), cbar_pad=0.0, cbar_shrink=1)
    plt.show()


if True:
    os.system(f'cosmoglobe gnom {path}{dust} -lon 30 -lat 70 -comp freqmap -ticks auto -freq 70 -show')
    os.system(f'cosmoglobe plot {path}{dust} -comp freqmap -range 0.5 -freq 70 -show')
    os.system(f'cosmoglobe trace {path}{chain} -labels "1 2 3 4" -dataset synch/beta_pixreg_val -show')

if True:
    #c = Chain(path+chain)
    #print(c["000001/tod/023-WMAP_K/chisq"])
    trace(path+chain, figsize=(10,2), sig=0, labels=["Reg1","Reg2","Reg3","Reg4"], dataset="synch/beta_pixreg_val", subplot=(3,1,1), ylabel=r"$\beta_s^T$")
    trace(path+chain, sig=1, labels=["Reg1","Reg2","Reg3","Reg4"], dataset="synch/beta_pixreg_val", subplot=(3,1,2), ylabel=r"$\beta_s^P$")
    trace(path+chain, dataset="tod/023-WMAP_K/bp_delta", subplot=(3,1,3), ylabel=r"$\Delta_{bp}$")
    #trace(path+chain, dataset="tod/023-WMAP_K/chisq", subplot=(5,1,4), ylabel=r"$\chi^2$")
    #trace(path+chain, dataset="tod/023-WMAP_K/gain", subplot=(5,1,5), ylabel="gain", xlabel="Gibbs sample")
    plt.show()
    

if False:    
    standalone_colorbar("planck", ticks=[-0.1,0,0.1], unit=u.uK)
    plt.savefig(paperfigs+"colorbar_planck_pm0.1.pdf", bbox_inches = 'tight', pad_inches = 0)
    
    standalone_colorbar("planck", ticks=[-5,0,5], unit=u.uK)
    plt.savefig(paperfigs+"colorbar_planck_pm5.pdf", bbox_inches = 'tight', pad_inches = 0)

    standalone_colorbar("planck", ticks=[-3,0,3], unit=u.uK)
    plt.savefig(paperfigs+"colorbar_planck_pm3.pdf", bbox_inches = 'tight', pad_inches = 0)

    standalone_colorbar("planck", ticks=[-1,0,1], unit=u.uK)
    plt.savefig(paperfigs+"colorbar_planck_pm1.pdf", bbox_inches = 'tight', pad_inches = 0)

    standalone_colorbar("planck", ticks=[-10,0,10], unit=u.uK)
    plt.savefig(paperfigs+"colorbar_planck_pm10.pdf", bbox_inches = 'tight', pad_inches = 0)

"""
sky = model_from_chain(path+chain, nside=256,)
if False:
    for i, freq in enumerate([200,]): #enumerate([353,]):
        #outname = f'{str(int(freq)).zfill(4)}GHz.png'
        if freq > 180:
            s = f(freq)*1e-3
        elif freq < 20:
            s = f2(freq)
        else:
            s=1

        #s = 1/(f(freq)*1e-3) if freq > 180 else 1
        if freq < 353:
            if freq > 250:
                r = ((353-freq)/(353-250))*67
            else:
                r = 67

        else:
            r = 0
        cmb = hp.remove_dipole(sky(cmb, freq*u.GHz,)[0].value, gal_cut=30)
        fsky = hp.remove_dipole(sky(freq*u.GHz,)[0].value, gal_cut=30, copy=True)
        print(np.median(fsky))
        #fsky2 = transform(fsky[0])
        #plt.show()
        print(s,)
        plot(fsky, norm="log", remove_mono=True, cmap="planck", title=f'{freq:.0f} GHz',)
        plt.show()
        plot((fsky+1e6)*s, ticks=[-1e3, -100, 0, 100, 1e3, 1e7], norm="log", remove_mono=True, cmap="planck_log", title=f'{freq:.0f} GHz',)
        plt.show()
"""


from cosmoglobe.sky.components import AME, CMB, Dust, FreeFree, Radio, Synchrotron

COSMOGLOBE_COMPS = {
    comp.label: comp
    for comp in [
        AME,
        CMB,
        Dust,
        FreeFree,
        Radio,
        Synchrotron,
    ]
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cosmoglobe',
 'cosmoglobe.h5',
 'cosmoglobe.plot',
 'cosmoglobe.sky',
 'cosmoglobe.utils']

package_data = \
{'': ['*'], 'cosmoglobe': ['data/*']}

install_requires = \
['Pillow>=8.3.2',
 'astropy>=4.3.1,<5.0.0',
 'click>=8.0.1',
 'cmasher>=1.6.2,<2.0.0',
 'h5py>=3.0.0',
 'healpy>=1.15.0,<2.0.0',
 'matplotlib>=3.4.3,<4.0.0',
 'numba>=0.54.0,<0.55.0',
 'numpy>=1.17.0,<1.21.0',
 'rich>=10.9.0,<11.0.0',
 'scipy>=1.6.0',
 'tqdm>=4.62.2,<5.0.0']

entry_points = \
{'console_scripts': ['cosmoglobe = cosmoglobe.__main__:cli']}

setup_kwargs = {
    'name': 'cosmoglobe',
    'version': '0.9.38',
    'description': 'cosmoglobe is a python package that interfaces the Cosmoglobe Sky Model with commander3 outputs for the purpose of producing astrophysical sky maps.',
    'long_description': None,
    'author': 'Metin San',
    'author_email': 'metinisan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from typing import Dict, Any, Protocol

import astropy.units as u
import numpy as np

from cosmoglobe.h5.chain_contextfactory import ChainContextFactory


class ChainContext(Protocol):
    """Protocol defining context for chain files.

    Chain context defines additional processing required on the chain items
    before they are ready to be put into the sky model.
    """

    def __call__(self, args: Dict[str, Any]) -> Dict[str, Any]:
        """Function that performs the processing on the specific chain item.

        This function needs to manipulate and return the `args` dictionary
        """


class FreqRefContext:
    """Re-shapes freq_ref for unpolarized components."""

    def __call__(self, args: Dict[str, Any]) -> Dict[str, Any]:
        if "freq_ref" in args:
            args["freq_ref"] = args["freq_ref"].to("GHz")
            if args["amp"].shape[0] != 3:
                args["freq_ref"] = args["freq_ref"][0]

        return args


class RadioContext:
    """Context for the radio component in the chain.

    In the chain, the amp of radio is stored in units of 'mJy'. Here we
    manually set convert it to 'uK' through astropy.

    NOTE: we pretend that amp has units of 'mJy/sr' to be able convert
    the units properly. This is fine since we always end dividing by
    the beam_area when converting the amplitudes to HEALPIX maps.
    """

    def __call__(self, args: Dict[str, Any]) -> Dict[str, Any]:
        args["alpha"] = args["alpha"][0]
        args["amp"] = u.Quantity(args["amp"].value, unit="mJy/sr").to(
            u.uK, equivalencies=u.thermodynamic_temperature(args["freq_ref"])
        )

        return args


class MapToScalarContext:
    """Extract and returns a scalar.

    Datasets in the cosmoglobe chains tends to be stored in HEALPIX maps.
    A quantity is considered a scalar if it is constant over all axes of the
    dataset.
    """

    def __call__(self, args: Dict[str, Any]) -> Dict[str, Any]:
        for key, value in args.items():
            if np.size(value) > 1:
                if np.ndim(value) > 1:
                    uniques = [np.unique(col) for col in value]
                    all_cols_are_unique = all([len(col) == 1 for col in uniques])
                else:
                    uniques = np.unique(value)
                    all_cols_are_unique = len(uniques) == 1

                if all_cols_are_unique:
                    args[key] = u.Quantity(uniques, unit=value.unit)

        return args


chain_context = ChainContextFactory()

chain_context.register_context([], FreqRefContext)
chain_context.register_context([], MapToScalarContext)
chain_context.register_context(["radio"], RadioContext)

chain_context.register_mapping([], {"freq_ref": "nu_ref"})
chain_context.register_mapping(["radio"], {"alpha": "specind"})
chain_context.register_mapping(["ame"], {"freq_peak": "nu_p"})
chain_context.register_mapping(["ff"], {"T_e": "Te"})

chain_context.register_units([], {"amp": u.uK, "freq_ref": u.Hz})
chain_context.register_units(["ame"], {"nu_p": u.GHz})
chain_context.register_units(["dust"], {"T": u.K})
chain_context.register_units(["ff"], {"T_e": u.K})



from abc import ABC, abstractmethod
from typing import Dict, Optional, Union
import warnings

import astropy.units as u
import healpy as hp
import numpy as np

from cosmoglobe.utils.utils import gaussian_beam_2D, to_unit
from cosmoglobe.utils.bandpass import (
    get_bandpass_coefficient,
    get_bandpass_scaling,
    get_normalized_bandpass,
)

warnings.simplefilter("once", UserWarning)


class SkyComponent(ABC):
    """Abstract base class for a sky component.

    This class implements the `__call__` function which is responsible for
    simulating and returning the sky emission. Additionally, it enforces
    the implementation of a `_smooth_emission` function for any sub class.
    """

    def __init__(
        self,
        amp: u.Quantity,
        freq_ref: u.Quantity,
        **spectral_parameters: Dict[str, u.Quantity],
    ) -> None:
        """Initializes a sky component and reshape inputs."""

        self._amp = self._reshape_amp(amp)
        self._freq_ref = self._reshape_freq_ref(freq_ref)
        self._spectral_parameters = self._reshape_spectral_parameters(
            spectral_parameters
        )
        self._is_polarized = True if self._amp.shape[0] == 3 else False

    @property
    def amp(self) -> u.Quantity:
        """Amplitude map the sky component."""

        return self._amp

    @property
    def freq_ref(self) -> u.Quantity:
        """Reference frequencies of the amplitude map."""

        return self._freq_ref

    @property
    def spectral_parameters(self) -> Dict[str, u.Quantity]:
        """Spectral parameters."""

        return self._spectral_parameters

    @abstractmethod
    def _smooth_emission(self, emission: u.Quantity, fwhm: u.Quantity) -> u.Quantity:
        """Smooths simulated emission given a beam FWHM.

        Parameters
        ----------
        emission
            Simulated emission.
        fwhm
            The full width half max parameter of the Gaussian. Default: 0.0

        Returns
        -------
        emission
            Smoothed emission.
        """

    @u.quantity_input(
        freqs=u.Hz,
        bandpass=(u.Jy / u.sr, u.K, None),
        fwhm=(u.rad, u.deg, u.arcmin),
    )
    def __call__(
        self,
        freqs: u.Quantity,
        bandpass: Optional[u.Quantity] = None,
        fwhm: u.Quantity = 0.0 * u.rad,
        output_unit: Union[str, u.UnitBase] = u.uK,
    ) -> u.Quantity:
        r"""Simulates and returns the sky emission of a component.

        This method computes the sky emission of a component for a single
        frequency :math:`\nu` or integrated over a  bandpass :math:`\tau`.

        Parameters
        ----------
        freqs
            A frequency, or a list of frequencies for which to evaluate
            the sky emission.
        bandpass
            Bandpass profile corresponding to the frequencies. Default is
            None. If `bandpass` is None and `freqs` is a single frequency,
            a delta peak is assumed (unless `freqs` is a list of
            frequencies, for which a top-hat bandpass is used to perform
            bandpass integration instead).
        fwhm
            The full width half max parameter of the Gaussian (Default is
            0.0, which indicates no smoothing of output maps).
        output_unit
            The desired output units of the emission. The supported units are
            :math:`\mathrm{\mu K_{RJ}}` and :math:`\mathrm{MJ/sr}` (By
            default the output unit of the model is always in
            :math:`\mathrm{\mu K_{RJ}}`.

        Returns
        -------
        Emission
            The component emission.

        Notes
        -----
        This function computes the following expression for a given
        component:

        .. math::

            \boldsymbol{s}_i(\nu) = \boldsymbol{a}
            _i(\nu_{0,i}) \; f_{i}
            (\nu),

        where :math:`\boldsymbol{a}_i` is the amplitude map of
        component :math:`i` at some reference frequency
        :math:`\nu_{0,i}`, and :math:`f_i(\nu)` is the
        scaling factor given by the SED of the component.

        See Also
        --------
        Model.__call__
            See for explicit expressions for :math:`\boldsymbol{s}_i(\nu)`.

        Examples
        --------
        Simulated dust emission at :math:`350\; \mathrm{GHz}`:

        >>> from cosmoglobe import skymodel
        >>> import astropy.units as u
        >>> model = skymodel(nside=256)
        >>> model.dust(350*u.GHz)[0]
        [21.12408523 -0.30044442  5.8907252  ...  4.6465226   6.94981578
          8.23490773] uK

        Simulated synchrotron emission at :math:`500\; \mathrm{GHz}`
        smoothed with a :math:`50\; '` Gaussian beam, outputed in units of
        :math:`\mathrm{MJy} / \mathrm{sr}`:

        >>> model.synch(40*u.GHz, fwhm=50*u.arcmin, output_unit='MJy/sr')[0]
        [0.00054551 0.00053428 0.00051471 ... 0.00046559 0.00047185 0.00047065]
         MJy / sr
        """

        if np.size(freqs) == 1:
            emission = self._get_delta_emission(freqs)
            emission = self._smooth_emission(emission, fwhm)
            if output_unit is not None:
                emission = to_unit(emission, freqs, output_unit)
        else:
            emission = self._get_bandpass_emission(
                freqs, bandpass, output_unit=output_unit
            )
            emission = self._smooth_emission(emission, fwhm)

        return emission

    def _get_delta_emission(self, freq: u.Quantity) -> u.Quantity:
        """Simulates the component emission at a delta frequency."""

        scaling = self._get_freq_scaling(
            freq, self.freq_ref, **self.spectral_parameters
        )

        return self.amp * scaling

    def _get_bandpass_emission(
        self,
        freqs: u.Quantity,
        bandpass: u.Quantity = None,
        output_unit: u.UnitBase = u.uK,
    ) -> u.Quantity:
        """Simulates the component emission at over a bandpass."""

        if bandpass is None:
            warnings.warn("Bandpass is None. Defaulting to top-hat bandpass")
            bandpass = np.ones(freqs_len := len(freqs)) / freqs_len * u.K

        bandpass = get_normalized_bandpass(bandpass, freqs)
        bandpass_coefficient = get_bandpass_coefficient(bandpass, freqs, output_unit)

        bandpass_scaling = get_bandpass_scaling(freqs, bandpass, self)
        emission = self.amp * bandpass_scaling * bandpass_coefficient

        return emission

    @staticmethod
    def _reshape_amp(amp: u.Quantity) -> u.Quantity:
        """Reshapes the reference frequency to a broadcastable shape."""

        return amp if amp.ndim != 1 else np.expand_dims(amp, axis=0)

    @staticmethod
    def _reshape_freq_ref(freq_ref: u.Quantity) -> u.Quantity:
        """Reshapes the reference frequency to a broadcastable shape."""

        if freq_ref is None:
            return
        elif freq_ref.size == 1:
            if np.ndim(freq_ref) == 0:
                return np.expand_dims(freq_ref, axis=0)
            return freq_ref
        elif freq_ref.size == 2:
            return np.expand_dims(
                u.Quantity([freq_ref[0], freq_ref[1], freq_ref[1]]), axis=1
            )
        elif freq_ref.size == 3:
            return freq_ref.reshape((3, 1))
        else:
            raise ValueError("Unrecognized shape")

    @staticmethod
    def _reshape_spectral_parameters(
        spectral_parameters: Dict[str, u.Quantity]
    ) -> Dict[str, u.Quantity]:
        """Reshapes the spectral parameters to a broadcastable shape."""

        return {
            key: (np.expand_dims(value, axis=0) if value.ndim == 1 else value)
            for key, value in spectral_parameters.items()
        }

    def __repr__(self) -> str:
        """Representation of the component."""

        main_repr = f"{self.__class__.__name__}"
        main_repr += "("
        extra_repr = ""
        for key in self.spectral_parameters.keys():
            extra_repr += f"{key}, "
        if extra_repr:
            extra_repr = extra_repr[:-2]
        main_repr += extra_repr
        main_repr += ")"

        return main_repr


class Diffuse(SkyComponent):
    """Abstract base class for diffuse sky components."""

    @abstractmethod
    def _get_freq_scaling(
        freq: u.Quantity,
        freq_ref: u.Quantity,
        **spectral_parameters: Dict[str, u.Quantity],
    ) -> u.Quantity:
        r"""Computes the frequency scaling for a component.

        Parameters
        ----------
        freq
            Frequency at which to evaluate the model.
        freq_ref
            Reference frequencies for the components `amp` map.
        spectral_parameters
            The spectral parameters of the component.

        Returns
        -------
        scaling
            Frequency scaling factor with dimensionless units.
        """

    def _smooth_emission(self, emission: u.Quantity, fwhm: u.Quantity) -> u.Quantity:
        """See base class."""

        if fwhm == 0.0:
            return emission

        fwhm = (fwhm.to(u.rad)).value
        if self._is_polarized:
            emission = u.Quantity(hp.smoothing(emission, fwhm=fwhm), unit=emission.unit)
        else:
            emission[0] = u.Quantity(
                hp.smoothing(emission[0], fwhm=fwhm), unit=emission.unit
            )

        return emission


class PointSource(SkyComponent):
    """Abstract base class for point source sky components."""

    _nside: int = None

    @abstractmethod
    def _get_freq_scaling(
        freq: u.Quantity,
        freq_ref: u.Quantity,
        **spectral_parameters: Dict[str, u.Quantity],
    ) -> u.Quantity:
        r"""Computes the frequency scaling for a component.

        Parameters
        ----------
        freq
            Frequency at which to evaluate the model.
        freq_ref
            Reference frequencies for the components `amp` map.
        spectral_parameters
            The spectral parameters of the component.

        Returns
        -------
        scaling
            Frequency scaling factor with dimensionless units.
        """

    def _smooth_emission(self, emission: u.Quantity, fwhm: u.Quantity) -> u.Quantity:
        """See base class.

        In addition to smoothing the point sources, this method also maps
        the cataloged points to a healpix map in the process.
        """

        return self._points_to_map(emission, fwhm=fwhm)

    def _points_to_map(
        self, amp: u.Quantity, fwhm: u.Quantity, n_fwhm: int = 2, verbose: bool = False
    ) -> u.Quantity:
        """Maps the cataloged point sources onto a healpix map with a truncated
        gaussian beam. For more information, see
         `Mitra et al. (2010) <https://arxiv.org/pdf/1005.1929.pdf>`_.

        Parameters
        ----------
        amp
            Amplitudes of the point sources.
        fwhm
            The full width half max parameter of the Gaussian. Default: 0.0
        n_fwhm
            The fwhm multiplier used in computing radial cut off r_max
            calculated as r_max = n_fwhm * fwhm of the Gaussian.
            Default: 2.
        """

        nside = self._nside
        if nside is None:
            raise AttributeError(
                "The _nside attribute of a point source have not been set"
            )

        angular_coords = self.angular_coords

        if amp.ndim > 1:
            amp = np.squeeze(amp)

        healpix_map = u.Quantity(np.zeros(hp.nside2npix(nside)), unit=amp.unit)
        pix_lon, pix_lat = hp.pix2ang(
            nside, np.arange(hp.nside2npix(nside)), lonlat=True
        )

        fwhm = fwhm.to(u.rad)
        sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))

        if sigma == 0.0:
            # No smoothing nesecarry. We directly map the sources to pixels
            warnings.warn("Mapping point sources to pixels without beam smoothing.")
            pixels = hp.ang2pix(nside, *angular_coords.T, lonlat=True)
            beam_area = hp.nside2pixarea(nside) * u.sr
            healpix_map[pixels] = amp
            healpix_map /= beam_area.value

        else:
            # Apply a truncated gaussian beam to each point source
            pix_res = hp.nside2resol(nside)
            if fwhm.value < pix_res:
                raise ValueError(
                    "fwhm must be >= pixel resolution to resolve the " "point sources."
                )
            beam_area = 2 * np.pi * sigma ** 2
            r_max = n_fwhm * fwhm.value

            sigma = sigma.value
            beam = gaussian_beam_2D
            print("Smoothing point sources...")

            for idx, (lon, lat) in enumerate(angular_coords):
                vec = hp.ang2vec(lon, lat, lonlat=True)
                inds = hp.query_disc(nside, vec, r_max)
                r = hp.rotator.angdist(
                    np.array([pix_lon[inds], pix_lat[inds]]),
                    np.array([lon, lat]),
                    lonlat=True,
                )
                healpix_map[inds] += amp[idx] * beam(r, sigma)

        healpix_map /= beam_area.value

        return np.expand_dims(healpix_map, axis=0)

    def _read_coords_from_catalog(self, catalog: str) -> np.ndarray:
        """Reads in the angular coordinates of the point sources from a
        given catalog.

        Parameters
        ----------
        catalog
            Path to the point source catalog. Default is the COM_GB6
            catalog.

        Returns
        -------
        coords
            Longitude and latitude values of each point source
        """

        try:
            coords = np.loadtxt(catalog, usecols=(0, 1))
        except OSError:
            raise OSError("Could not find point source catalog")

        if len(coords) != len(self.amp[0]):
            raise ValueError("Cataloge does not match chain catalog")

        return coords



from pathlib import Path
import warnings

import astropy.units as u
import numpy as np
import healpy as hp

from cosmoglobe.sky.base import Diffuse, PointSource
from cosmoglobe.utils.functions import (
    thermodynamical_to_brightness,
    blackbody_emission,
    gaunt_factor,
)


DATA_DIR = Path(__file__).parent.parent.resolve() / "data"
RADIO_CATALOG = DATA_DIR / "radio_catalog.dat"
SPDUST2_FILE = DATA_DIR / "spdust2_cnm.dat"


class Synchrotron(Diffuse):
    r"""Class representing the synchrotron component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of synchrotron at the reference frequencies given
        by `freq_ref` from a commander3 chain.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,s}` for `amp`.
    spectral_parameters : dict
        Dictionary containing the power law spectral index :math:`\beta`.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    This is a generic power law given at a reference frequency
    :math:`\nu_{s,0}` with a power law :math:`\beta` in Rayleigh-Jeans
    temperature. It is defined using the convention in
    `BeyondPlanck (2020), Section 3.3.1
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{s}}(\nu) \propto
        \left( \frac{\nu}{\nu_\mathrm{0,s}} \right)^
        {\beta + C \ln \nu / \nu_{0,s}},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission. :math:`C` is set to 0 for all current implementations
    as of BP9.
    """

    label = "synch"

    def __init__(self, amp: u.Quantity, freq_ref: u.Quantity, beta: u.Quantity) -> None:
        super().__init__(amp, freq_ref, beta=beta)

    def _get_freq_scaling(
        self, freq: u.Quantity, freq_ref: u.Quantity, beta: u.Quantity
    ) -> u.Quantity:
        r"""See base class.

        Parameters
        ----------
        beta
            Power law spectral index.
        """

        scaling = (freq / freq_ref) ** beta

        return scaling


class Dust(Diffuse):
    r"""Class representing the thermal dust component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of thermal dust at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,d}` for `amp`.
    spectral_parameters : dict
        Dictionary containing the spectral parameters :math:`\beta` and
        :math:`T`.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    This is a modified blackbody with a power law spectral index
    :math:`\beta_\mathrm{d}` in Rayleigh-Jeans temperature, and thermal dust
    temperature :math:`T_{\mathrm{d}}`. It is defined using the convention
    in `BeyondPlanck (2020), Section 3.3.3
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{d}}(\nu) \propto
        \frac{\nu^{\beta_{\mathrm{d}}+1}}{\mathrm{e}^
        {h\nu/kT_{\mathrm{d}}}-1},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`h` is Planck's constant, and :math:`k` is the
    Boltzmann constant.
    """

    label = "dust"

    def __init__(
        self, amp: u.Quantity, freq_ref: u.Quantity, beta: u.Quantity, T: u.Quantity
    ) -> None:
        super().__init__(amp, freq_ref, beta=beta, T=T)

    def _get_freq_scaling(
        self, freq: u.Quantity, freq_ref: u.Quantity, beta: u.Quantity, T: u.Quantity
    ) -> u.Quantity:
        r"""See base class.

        Parameters
        ----------
        beta
            Power law spectral index.
        T
            Temperature of the blackbody.
        """

        blackbody_ratio = blackbody_emission(freq, T) / blackbody_emission(freq_ref, T)
        scaling = (freq / freq_ref) ** (beta - 2) * blackbody_ratio

        return scaling


class FreeFree(Diffuse):
    r"""Class representing the free-free component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of free-free at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,\mathrm{ff}}` for `amp`.
    spectral_parameters : dict
        Dictionary containing the spectral parameter :math:`T_e`.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    The free-free emission is defined using the convention in
    `BeyondPlanck (2020), Section 3.3.2
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{ff}}(\nu) \propto
        \frac{g_{\mathrm{ff}}(T_\mathrm{e})}{\nu^2},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`g_\mathrm{ff}` is the Gaunt factor, and
    :math:`T_\mathrm{e}` is the electron temperature.
    """

    label = "ff"

    def __init__(self, amp: u.Quantity, freq_ref: u.Quantity, T_e: u.Quantity) -> None:
        super().__init__(amp, freq_ref, T_e=T_e)

    def _get_freq_scaling(
        self, freq: u.Quantity, freq_ref: u.Quantity, T_e: u.Quantity
    ) -> u.Quantity:
        r"""See base class.

        Parameters
        ----------
        T_e
            Electron temperature.
        """

        gaunt_factor_ratio = gaunt_factor(freq, T_e) / gaunt_factor(freq_ref, T_e)
        scaling = (freq_ref / freq) ** 2 * gaunt_factor_ratio

        return scaling


class AME(Diffuse):
    r"""Class representing the spinning dust component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission maps of spinning dust at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,sd}` for `amps`.
    spectral_parameters : dict
        Dictionary containing the spectral parameter :math:`\nu_p`.
    label : str
        Component label.
    spdust2 : `astropy.units.Quantity`
        spdust2 template.

    Methods
    -------
    __call__

    Notes
    -----
    The spinning dust emission is defined using the convention in
    `BeyondPlanck (2020), Section 3.3.4
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{sd}}(\nu) \propto
        \nu^{-2} \boldsymbol{s}_{0}^{\mathrm{sd}}\left(\nu \cdot
        \frac{30.0\; \mathrm{GHz}}{\nu_{p}}\right)

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`\boldsymbol{s}_0^{\mathrm{sd}}` is the `spdust2`
    template, and :math:`\nu_p` the peak frequency.
    """

    label = "ame"

    def __init__(
        self, amp: u.Quantity, freq_ref: u.Quantity, freq_peak: u.Quantity
    ) -> None:
        super().__init__(amp, freq_ref, freq_peak=freq_peak)

        spdust2_freq, spdust2_amp = np.loadtxt(SPDUST2_FILE, unpack=True)
        spdust2_freq = u.Quantity(spdust2_freq, unit=u.GHz)
        spdust2_amp = u.Quantity(spdust2_amp, unit=(u.Jy / u.sr))
        spdust2_amp = spdust2_amp.to(
            u.K, equivalencies=u.brightness_temperature(spdust2_freq)
        )
        self.spdust2 = np.array([spdust2_freq.si.value, spdust2_amp.si.value])

    def _get_freq_scaling(
        self, freq: u.Quantity, freq_ref: u.Quantity, freq_peak: u.Quantity
    ) -> u.Quantity:
        r"""See base class.

        Parameters
        ----------
        freq_peak
            Peak frequency.
        """
        spdust2 = self.spdust2
        peak_scale = 30 * u.GHz / freq_peak

        # AME is undefined at outside of this frequency range
        if not np.logical_and(
            (freq * peak_scale).si.value > np.min(spdust2[0]),
            (freq * peak_scale).si.value < np.max(spdust2[0]),
        ).all():
            return u.Quantity(0, unit=u.dimensionless_unscaled)

        interp = np.interp((freq * peak_scale).si.value, spdust2[0], spdust2[1])
        interp_ref = np.interp((freq_ref * peak_scale).si.value, spdust2[0], spdust2[1])
        scaling = interp / interp_ref

        return scaling


class CMB(Diffuse):
    r"""Class representing the CMB component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Emission map of CMB in units of :math:`\mathrm{\mu K_{CMB}}`
    freq_ref : `astropy.units.Quantity`
        Reference frequency is None due to the amplitude maps being
        stored in units of :math:`\mathrm{\mu K_{CMB}}`.
    spectral_parameters : dict
        Empty dictionary.
    label : str
        Component label.

    Methods
    -------
    __call__
    remove_dipole
    get_dipole

    Notes
    -----
    The CMB emission is defined using the convention in
    `BeyondPlanck (2020), Section 3.2
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{CMB}}(\nu) \propto
        \frac{x^{2} \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)
        ^{2}} \boldsymbol{s}^{\mathrm{CMB}},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`x=h \nu / k T_{0}` and
    :math:`T_0 = 2.7255 \mathrm{K}` as of BP9.
    """

    label = "cmb"

    def __init__(self, amp: u.Quantity, freq_ref: u.Quantity = None) -> None:
        super().__init__(amp, freq_ref=freq_ref)

    @u.quantity_input(gal_cut=u.deg)
    def get_dipole(self, gal_cut: u.Quantity = 10 * u.deg) -> u.Quantity:
        """Returns the solar dipole from the reference amplitude map.

        Parameters
        ----------
        gal_cut
            Masks pixles :math:`\pm` `gal_cut` in latitude before estimating
            dipole (:math:`10^\circ` by default).

        Returns
        -------
        dipole
            Map of the solar dipole.
        """

        if hasattr(self, "dipole"):
            warnings.warn(
                "Returning previously removed dipole signal (overriding gal_cut)."
            )
            return self.dipole

        amp_without_dipole = u.Quantity(
            hp.remove_dipole(self.amp[0], gal_cut=gal_cut.value), unit=self.amp.unit
        )

        return self.amp[0] - amp_without_dipole

    @u.quantity_input(gal_cut=u.deg)
    def remove_dipole(self, gal_cut: u.Quantity = 10 * u.deg) -> None:
        """Removes the solar dipole from the reference amplitude map.

        Parameters
        ----------
        gal_cut
            Masks pixles :math:`\pm` `gal_cut` in latitude before
            estimating dipole (:math:`10^\circ` by default).
        """

        amp_without_dipole = u.Quantity(
            hp.remove_dipole(self.amp[0], gal_cut=gal_cut.value), unit=self.amp.unit
        )

        self.dipole = self.amp[0] - amp_without_dipole
        self.amp[0] = amp_without_dipole

    def _get_freq_scaling(self, freq: u.Quantity, freq_ref=None) -> u.Quantity:
        r"""See base class."""

        # We explicitly expand the dims to support broadcasting
        return np.expand_dims(thermodynamical_to_brightness(freq), axis=0)


class Radio(PointSource):
    r"""Class representing the radio component.

    Attributes
    ----------
    amp : `astropy.units.Quantity`
        Point source amplitudes at the reference frequencies given
        by `freq_ref`. Note that this quantity is not a healpix map.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,\mathrm{src}}` for
        `amp`.
    spectral_parameters : dict
        Dictionary containing the spectral parameter :math:`\alpha`.
    nside : int
        Point source components need to be explicitly passed an NSIDE.
    label : str
        Component label.

    Methods
    -------
    __call__

    Notes
    -----
    This is a generic power law given at a reference frequency
    :math:`\nu_{\mathrm{0, src}}` with a power law spectral index
    :math:`\alpha`. It is defined using the convention in
    `BeyondPlanck (2020), Section 3.4.1
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{src}}(\nu) \propto
        \left(\frac{\nu}{\nu_{\mathrm{0, src}}}\right)^{\alpha-2}

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission.
    """

    label = "radio"

    def __init__(
        self, amp: u.Quantity, freq_ref: u.Quantity, alpha: u.Quantity
    ) -> None:
        super().__init__(amp, freq_ref, alpha=alpha)
        self.angular_coords = self._read_coords_from_catalog(RADIO_CATALOG)

    def _get_freq_scaling(
        self, freq: u.Quantity, freq_ref: u.Quantity, alpha: u.Quantity
    ) -> u.Quantity:
        r"""See base class.

        Parameters
        ----------
        specind
            Power law spectral index.
        """

        scaling = (freq / freq_ref) ** (alpha - 2)

        return scaling



# This file contains useful functions used across different plotting scripts.
from re import A, T
import warnings
from .. import data as data_dir
from cosmoglobe.sky.model import Model
from cosmoglobe.h5.model import model_from_chain
from cosmoglobe.h5.chain import Chain

import cmasher
from rich import print
import numpy as np
import healpy as hp
import matplotlib.pyplot as plt
import astropy.units as u
from matplotlib.colors import colorConverter, LinearSegmentedColormap, ListedColormap
from matplotlib.patches import Polygon
from matplotlib import _pylab_helpers
from pathlib import Path
import json

DEFAULT_FONTSIZES = {
    "xlabel": 11,
    "ylabel": 11,
    "xtick_label": 8,
    "ytick_label": 8,
    "title": 12,
    "cbar_label": 11,
    "cbar_tick_label": 9,
    "left_label": 11,
    "right_label": 11,
}
FIGURE_WIDTHS = {
    "x": 2.75,
    "s": 3.5,
    "m": 4.7,
    "l": 7,
}
STOKES = [
    "I",
    "Q",
    "U",
]


def set_style(darkmode=False):
    """
    This function sets the color parameter and text style
    """
    plt.rc(
        "font",
        family="serif",
    )
    plt.rcParams["mathtext.fontset"] = "stix"
    plt.rc(
        "text.latex",
        preamble=r"\usepackage{sfmath}",
    )

    tex_fonts = {
        # Use 10pt font in plots, to match 10pt font in document
        "axes.labelsize": 11,
        "font.size": 11,
        # Make the legend/label fonts a little smaller
        "legend.fontsize": 8,
        "xtick.labelsize": 8,
        "ytick.labelsize": 8,
    }
    plt.rcParams.update(
        tex_fonts,
    )

    ## If darkmode plot (white frames), change colors:
    params = [
        "text.color",
        "axes.edgecolor",
        "axes.labelcolor",
        "xtick.color",
        "ytick.color",
        "grid.color",
        "legend.facecolor",
        "legend.edgecolor",
    ]
    if darkmode:
        for p in params:
            plt.rcParams[p] = "white"
    else:
        for p in params:
            plt.rcParams[p] = "black"


def make_fig(
    figsize, fignum, hold, subplot, reuse_axes, darkmode=False, projection=None
):
    """
    Create matplotlib figure, add subplot, use current axes etc.
    """

    if figsize is None:
        fig_width_pt = 246.0  # Get this from LaTeX using \showthe\columnwidth
        inches_per_pt = 1.0 / 72.27  # Convert pt to inch
        golden_mean = (np.sqrt(5) - 1.0) / 2.0  # Aesthetic ratio
        fig_width = fig_width_pt * inches_per_pt  # width in inches
        fig_height = fig_width * golden_mean  # height in inches
        figsize = [fig_width, fig_height]

    #  From healpy
    nrows, ncols, idx = (1, 1, 1)
    width, height = figsize
    if not (hold or subplot or reuse_axes):
        # Set general style parameters
        set_style(darkmode)

        fig = plt.figure(fignum, figsize=(width, height))
    elif hold:
        fig = plt.gcf()
        # left, bottom, right, top = np.array(fig.gca().get_position()).ravel()
        fig.delaxes(fig.gca())
    elif reuse_axes:
        fig = plt.gcf()
    else:  # using subplot syntax
        if hasattr(subplot, "__len__"):
            nrows, ncols, idx = subplot
        else:
            nrows, ncols, idx = subplot // 100, (subplot % 100) // 10, (subplot % 10)

        # If figure exists, get. If not, create.
        figManager = _pylab_helpers.Gcf.get_active()
        if figManager is not None:
            fig = figManager.canvas.figure
        else:
            # Set general style parameters
            set_style(darkmode)
            figsize = (width * ncols, height * nrows)
            fig = plt.figure(figsize=figsize)

        if idx < 1 or idx > ncols * nrows:
            raise ValueError(
                "Wrong values for subplot: %d, %d, %d" % (nrows, ncols, idx)
            )

    if reuse_axes:
        ax = fig.gca()
    else:
        ax = fig.add_subplot(int(f"{nrows}{ncols}{idx}"), projection=projection)

    return fig, ax


def get_percentile(m, percentile):
    """
    This function gets appropriate min and max
    values of the data for a given percentile
    """
    vmin = np.percentile(m, 100.0 - percentile)
    vmax = np.percentile(m, percentile)

    vmin = 0.0 if abs(vmin) < 1e-5 else vmin
    vmax = 0.0 if abs(vmax) < 1e-5 else vmax
    return [vmin, vmax]


def fmt(x, pos):
    """
    Format color bar labels
    """
    if abs(x) > 1e4 or (abs(x) <= 1e-3 and abs(x) > 0):
        a, b = f"{x:.2e}".split("e")
        b = int(b)
        if float(a) == 1.00:
            return r"$10^{" + str(b) + "}$"
        elif float(a) == -1.00:
            return r"$-10^{" + str(b) + "}$"
        else:
            return fr"${a} \cdot 10^{b}$"
    elif abs(x) > 1e1 or (float(abs(x))).is_integer():
        return fr"${int(x):d}$"
    else:
        x = round(x, 2)
        return fr"${x}$"


def format_list(items):
    return [fmt(i, 1) for i in items]


def symlog(m, linthresh=1.0):
    """
    This is the semi-logarithmic function used when logscale=True
    """
    # Extra fact of 2 ln 10 makes symlog(m) = m in linear regime
    m = m / linthresh / (2 * np.log(10))
    return np.log10(0.5 * (m + np.sqrt(4.0 + m * m)))


def legend_positions(input):
    """
    Calculate position of labels to the right in plot...
    """
    positions = input[-1, :]

    def push(dpush):
        """
        ...by puting them to the last y value and
        pushing until no overlap
        """
        collisions = 0
        for column1, value1 in enumerate(positions):
            for column2, value2 in enumerate(positions):
                if column1 != column2:
                    dist = abs(value1 - value2)
                    if dist < dpush * 10:
                        collisions += 1
                        if value1 < value2:
                            positions[column1] -= dpush
                            positions[column2] += dpush
                        else:
                            positions[column1] += dpush
                            positions[column2] -= dpush
                            return True

    min_ = np.min(np.min(input))
    max_ = np.max(np.max(input))
    dpush = np.abs(max_ - min_) * 0.01

    pushings = 0
    while True:
        if pushings == 1000:
            dpush *= 10
            pushings = 0
        pushed = push(dpush)
        if not pushed:
            break
        if pushings > 100:
            break
        pushings += 1
    return positions


def gradient_fill(x, y, fill_color=None, ax=None, alpha=1.0, invert=False, **kwargs):
    """
    Plot a line with a linear alpha gradient filled beneath it.

    Parameters
    ----------
    x, y : array-like
        The data values of the line.
    fill_color : a matplotlib color specifier (string, tuple) or None
        The color for the fill. If None, the color of the line will be used.
    ax : a matplotlib Axes instance
        The axes to plot on. If None, the current pyplot axes will be used.
    Additional arguments are passed on to matplotlib's ``plot`` function.

    Returns
    -------
    line : a Line2D instance
        The line plotted.
    im : an AxesImage instance
        The transparent gradient clipped to just the area beneath the curve.
    """

    if ax is None:
        ax = plt.gca()

    (line,) = ax.plot(x, y, **kwargs)
    if fill_color is None:
        fill_color = line.get_color()

    zorder = line.get_zorder()
    # alpha = line.get_alpha()
    alpha = 1.0 if alpha is None else alpha

    z = np.empty((100, 1, 4), dtype=float)
    rgb = colorConverter.to_rgb(fill_color)
    z[:, :, :3] = rgb
    z[:, :, -1] = np.linspace(0, alpha, 100)[:, None]

    xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
    if invert:
        ymin, ymax = (ymax, ymin)
    im = ax.imshow(
        z, aspect="auto", extent=[xmin, xmax, ymin, ymax], origin="lower", zorder=zorder
    )

    xy = np.column_stack([x, y])
    xy = np.vstack([[xmin, ymin], xy, [xmax, ymin], [xmin, ymin]])
    clip_path = Polygon(xy, facecolor="none", edgecolor="none", closed=True)
    ax.add_patch(clip_path)
    im.set_clip_path(clip_path)

    ax.autoscale(True)
    return line, im


def gradient_fill_between(ax, x, y1, y2, color="#ffa15a", alpha=0.5):
    N = 100
    y = np.zeros((N, len(y1)))
    for i in range(len(y1)):
        y[:, i] = np.linspace(y1[i], y2[i], N)

    alpha = np.linspace(0, alpha, 100) ** 1.5
    for i in range(1, N):
        ax.fill_between(x, y[i - 1], y[i], color=color, alpha=alpha[i], zorder=-10)
        ax.set_rasterization_zorder(-1)


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx], idx


def standalone_colorbar(
    cmap,
    ticks,
    ticklabels=None,
    unit=None,
    fontsize=None,
    norm=None,
    shrink=0.3,
    darkmode=False,
    width=10,
):

    set_style(darkmode)
    if fontsize is None:
        fontsize = DEFAULT_FONTSIZES
    plt.figure(figsize=(width, 2))
    img = plt.imshow(np.array([[ticks[0], ticks[-1]]]), load_cmap(cmap))
    plt.gca().set_visible(False)
    if ticklabels is None:
        ticklabels = format_list(ticks)
    apply_colorbar(
        plt.gcf(),
        plt.gca(),
        img,
        ticks,
        ticklabels,
        unit,
        fontsize=fontsize,
        norm=norm,
        cbar_shrink=shrink,
        cbar_pad=0.0
    )


def apply_logscale(m, ticks, linthresh=1):
    """
    This function converts the data to logscale,
    and also converts the tick locations correspondingly
    NOTE: This should ideally be done for the colorbar, not the data.
    """

    print("[magenta]Applying semi-logscale[/magenta]")
    m = symlog(m, linthresh)
    new_ticks = []
    for i in ticks:
        new_ticks.append(symlog(i, linthresh))

    m = np.maximum(np.minimum(m, new_ticks[-1]), new_ticks[0])
    return m, new_ticks


def apply_colorbar(
    fig,
    ax,
    image,
    ticks,
    ticklabels,
    unit,
    fontsize=None,
    linthresh=1,
    norm=None,
    cbar_pad=0.04,
    cbar_shrink=0.3,
):
    """
    This function applies a colorbar to the figure and formats the ticks.
    """
    from matplotlib.ticker import FuncFormatter

    if fontsize is None:
        fontsize = DEFAULT_FONTSIZES
    cb = fig.colorbar(
        image,
        ax=ax,
        orientation="horizontal",
        shrink=cbar_shrink,
        pad=cbar_pad,
        ticks=ticks,
        format=FuncFormatter(fmt),
    )
    cb.ax.set_xticklabels(ticklabels, size=fontsize["cbar_tick_label"])
    if isinstance(unit, u.UnitBase):
        unit = unit.to_string("latex")
    cb.ax.xaxis.set_label_text(unit, size=fontsize["cbar_label"])
    if norm == "log":
        linticks = np.linspace(-1, 1, 3) * linthresh
        logmin = np.round(ticks[0])
        logmax = np.round(ticks[-1])

        logticks_min = -(10 ** np.arange(0, abs(logmin) + 1))
        logticks_max = 10 ** np.arange(0, logmax + 1)
        ticks_ = np.unique(np.concatenate((logticks_min, linticks, logticks_max)))
        # cb.set_ticks(np.concatenate((ticks,symlog(ticks_))), []) # Set major ticks

        logticks = symlog(ticks_, linthresh)
        logticks = [x for x in logticks if x not in ticks]
        cb.set_ticks(np.concatenate((ticks, logticks)))  # Set major ticks
        cb.ax.set_xticklabels(ticklabels + [""] * len(logticks))

        minorticks = np.linspace(-linthresh, linthresh, 5)
        minorticks2 = np.arange(2, 10) * linthresh

        for i in range(len(logticks_min)):
            minorticks = np.concatenate((-(10 ** i) * minorticks2, minorticks))
        for i in range(len(logticks_max)):
            minorticks = np.concatenate((minorticks, 10 ** i * minorticks2))

        minorticks = symlog(minorticks, linthresh)
        minorticks = minorticks[(minorticks >= ticks[0]) & (minorticks <= ticks[-1])]
        cb.ax.xaxis.set_ticks(minorticks, minor=True)

    cb.ax.tick_params(
        which="both",
        axis="x",
        direction="in",
    )
    cb.ax.xaxis.labelpad = 0
    # workaround for issue with viewers, see colorbar docstring
    cb.solids.set_edgecolor("face")
    return cb


def load_cmap(cmap):
    """
    This function takes the colormap label and loads the colormap object
    """

    if cmap is None:
        cmap = "planck"
    if "planck" in cmap:
        if "planck_log" in cmap:  # logscale:
            # setup nonlinear colormap

            class GlogColormap(LinearSegmentedColormap):
                name = cmap

                def __init__(self, cmap, vmin=-1e3, vmax=1e7):
                    self.cmap = cmap
                    self.N = self.cmap.N
                    self.vmin = vmin
                    self.vmax = vmax

                def is_gray(self):
                    return False

                def __call__(self, xi, alpha=1.0, **kw):
                    x = xi * (self.vmax - self.vmin) + self.vmin
                    yi = self.modsinh(x)
                    # range 0-1
                    yi = (yi + 3) / 10.0
                    return self.cmap(yi, alpha)

                def modsinh(self, x):
                    return np.log10(0.5 * (x + np.sqrt(x ** 2 + 4)))

            cmap_path = Path(data_dir.__path__[0]) / "planck_cmap_logscale.dat"
            planck_cmap = np.loadtxt(cmap_path) / 255.0
            cmap = ListedColormap(planck_cmap, "planck")
            cmap = GlogColormap(cmap)
        else:
            cmap_path = Path(data_dir.__path__[0]) / "planck_cmap.dat"
            planck_cmap = np.loadtxt(cmap_path) / 255.0
            if cmap.endswith("_r"):
                planck_cmap = planck_cmap[::-1]
            cmap = ListedColormap(planck_cmap, "planck")
    elif "wmap" in cmap:
        cmap_path = Path(data_dir.__path__[0]) / "wmap_cmap.dat"
        wmap_cmap = np.loadtxt(cmap_path) / 255.0
        if cmap.endswith("_r"):
            wmap_cmap = wmap_cmap[::-1]
        cmap = ListedColormap(wmap_cmap, "wmap")
    elif cmap.startswith("black2"):
        cmap = LinearSegmentedColormap.from_list(cmap, cmap.split("2"))
    else:
        try:
            cmap = eval(f"cmasher.{cmap}")
        except:
            try:

                cmap = eval(f"cm.{cmap}")
            except:

                cmap = plt.get_cmap(cmap)

    # print("Colormap:" + f" {cmap.name}")
    return cmap


def get_data(input, sig, comp, freq, fwhm, nside=None, sample=-1):
    # Parsing component string
    fwhm_ = None
    specparam = "amp"
    if comp is not None:
        comp, *specparam_ = comp.split()
        if specparam_:
            specparam = specparam_[0]

    # If map is string, read data
    if isinstance(input, str):
        # If string input, then read either hdf or fits
        if input.endswith(".h5"):
            if comp is None:
                if freq is None:
                    print(
                        "[bold magenta]Warning! Neither frequency nor component selected. Plotting sky at 70GHz[/bold magenta]"
                    )
                data = model_from_chain(
                    input, components=comp, nside=nside, samples=sample
                )
            else:
                if freq is not None:
                    # If frequency is specified, simulate sky with model.
                    data = model_from_chain(
                        input, components=comp, nside=nside, samples=sample
                    )
                else:
                    chain = Chain(
                        input,
                    )
                    lmax = chain.get(f"{comp}/{specparam}_lmax", samples=sample)
                    alms = chain.get(f"{comp}/{specparam}_alm", samples=sample)
                    fwhm_ = chain.parameters[comp]["fwhm"] * u.arcmin
                    if nside is None: nside = chain.parameters[comp]["nside"]
                    pol = True if specparam == "amp" and alms.shape[0] == 3 else False
                    # PROBLEM! Get issue with nside when udgrading directly to low nside
                    data = hp.alm2map(
                        alms,
                        nside=nside,
                        fwhm=fwhm_.to(u.rad).value,
                        lmax=lmax,
                        mmax=lmax,
                        pol=pol,
                    )

        elif input.endswith(".fits"):
            data = hp.read_map(input, field=sig)
        else:
            raise ValueError("Input file must be .fits or .h5")
    else:
        data = input

    if isinstance(data, Model):
        """
        Get data from model object with frequency scaling
        """
        if comp is None:
            if freq is None:
                print(
                    "[bold magenta]Warning! Neither frequency nor component selected. Plotting sky at 70GHz[/bold magenta]"
                )
                freq = 70 * u.GHz
            comp = "freqmap"
            m = data(freq, fwhm=fwhm)
        else:
            if specparam == "amp":
                if freq is None:
                    # Get map at reference frequency if freq not specified
                    freq_ref = getattr(data, comp).freq_ref
                    freq = freq_ref.value
                    m = getattr(data, comp).amp
                    if comp == "radio":
                        m = getattr(data, comp)(freq_ref, fwhm=fwhm)
                    try:
                        freq = round(freq.squeeze()[sig], 5) * freq_ref.unit
                    except IndexError:
                        freq = round(freq, 5) * freq_ref.unit
                else:
                    # If freq is specified, scale with sky model
                    m = getattr(data, comp)(freq, fwhm=fwhm)
            else:
                m = getattr(data, comp).spectral_parameters[specparam]

                if len(m[sig]) == 1:
                    warnings.warn(
                        "Same value across the whole sky, mapping to array of length Npix"
                    )
                    m = np.full(hp.nside2npix(data.nside), m[sig])

    elif isinstance(data, np.ndarray):
        m = data
        if m.ndim == 1 and sig > 0 and not isinstance(input, str):
            print(
                f"[magenta]Input array is 1d, specified signal will only be used for labeling, make sure this is correct.[/magenta]"
            )
    else:
        raise TypeError(
            f"Type {type(data)} of data not supported"
            f"Supports numpy array, cosmoglobe model object or fits file string"
        )

    # Make sure it is a 1d array
    if m.ndim > 1:
        m = m[sig]

    # Convert astropy map to numpy array
    if isinstance(m, u.Quantity):
        m = m.value

    # ud_grade map
    if nside is not None and nside != hp.get_nside(m):
        print("Ud_grading map")
        m = hp.ud_grade(m, nside)

    # Smooth map
    if fwhm > 0.0 and fwhm_ is None:
        m = hp.smoothing(m, fwhm.to(u.rad).value)

    return m, comp, freq, hp.get_nside(m)


def get_params(**params):
    """
    This parses the autoparams json file for automatically setting parameters
    for an identified sky component.
    """
    freq = params["freq_ref"]
    # Convert astropy unit to latex
    if isinstance(params["unit"], u.UnitBase):
        params["unit"] = params["unit"].to_string("latex")

    # Autoset plotting parameters if component is specified
    if params["comp"] is not None:
        with open(Path(data_dir.__path__[0]) / "autoparams.json", "r") as f:
            autoparams = json.load(f)
        try:
            autoparams = autoparams[params["comp"]]
        except KeyError:
            print(
                f"[magenta]Component label {params['comp']} not found. Using unidentified profile.[/magenta]"
            )
            print(f"available keys are: {autoparams.keys()}")
            autoparams = autoparams["unidentified"]

        # Replace default values if None
        params = {
            key: (autoparams[key] if key in autoparams and value is None else value)
            for key, value in params.items()
        }

        # NOTE: This is really ugly code for selecting pol parameters
        # Pick polarization or temperature
        l = 0 if params["sig"] < 1 else -1
        for key in ["cmap", "ticks", "freq_ref"]:
            if isinstance(params[key], list):
                if key == "ticks":
                    if isinstance(params[key][0], list):
                        params[key] = params[key][l]
                else:
                    params[key] = params[key][l]

        if params["freq_ref"] is not None:
            params["freq_ref"] * u.GHz

        if params["left_label"] is None:
            params["left_label"] = ["I", "Q", "U"][params["sig"]]
        specials = ["residual", "freqmap", "bpcorr", "smap"]
        if any(j in params["comp"] for j in specials):
            if freq is not None:
                params["right_label"] = (
                    params["right_label"]
                    + "{"
                    + f'{("%.5f" % freq.value).rstrip("0").rstrip(".")}'
                    + "}"
                )
            else:
                params["right_label"] = None
                warnings.warn(
                    f'Specify frequency with -freq for automatic frequency labeling with the "freqmap" profile'
                )

        """
        # NOTE: This probably doesnt work with the current state of the comp
        # keyword. Rewrite somehow.
        if "rms" in params["comp"]:
            params["right_label"] += "^{\mathrm{RMS}}"
            params["cmap"] = "neutral"
            params["ticks"] = "comp"
        if "stddev" in params["comp"]:
            params["right_label"] = "\sigma_{\mathrm{" + params["right_label"] + "}}"
            params["cmap"] = "neutral"
            params["ticks"] = "comp"
        if "mean" in params["comp"]:
            params["right_label"] = "\langle " + params["right_label"] + "\rangle"
        """

        # Specify at which frequency we are observing in unit lable
        if freq is not None and params["unit"] is not None:
            params["unit"] = (
                f'{params["unit"]}\,@\,{("%.5f" % freq.value).rstrip("0").rstrip(".")}'
                + "\,\mathrm{GHz}"
            )

        # In the case that a difference frequency than the reference
        # is requested for a component map.
        if params["ticks"] == None:
            if (
                freq is not None
                and params["freq_ref"] != freq.value
                and params["comp"] not in specials
            ):
                warnings.warn(
                    f"Input frequency is different from reference, autosetting ticks"
                )
                print(f'input: {freq}, reference: {params["freq_ref"]}')
                params["ticks"] = "auto"

    # If ticks is not set automatically, try inputs
    if params["ticks"] is None:
        params["ticks"] = [params["min"], params["max"]]

    # if ticks is not specified to be auto in function call
    if params["ticks"] != "auto":
        # If range is not specified in function call
        if params["rng"] is not None:
            params["ticks"] = [-params["rng"], 0.0, params["rng"]]
        else:
            # If min is not specified in function call
            if params["min"] is not None:
                # If ticks is set to be automatic from autoparams
                if params["ticks"] == "auto":
                    params["ticks"] = [None, None]  # (These will be set later)
                params["ticks"][0] = params["min"]
            # If max is not specified in function call
            if params["max"] is not None:
                # If tick is set to be automatic from atoparams
                if params["ticks"] == "auto":
                    params["ticks"] = [None, None]  # (These will be set later)
                params["ticks"][-1] = params["max"]

    # Ticks and ticklabels
    if params["ticks"] == "auto" or params["norm"] == "hist":
        params["ticks"] = get_percentile(params["data"], 97.5)
    elif None in params["ticks"]:
        pmin, pmax = get_percentile(params["data"], 97.5)
        if params["ticks"][0] is None:
            params["ticks"][0] = pmin
        if params["ticks"][-1] is None:
            params["ticks"][-1] = pmax

    # Math text in labels
    for key in ["right_label", "left_label", "unit"]:
        if params[key] and params[key] != "":
            if "$" not in params[key]:
                params[key] = r"$" + params[key] + "$"

    # Special case if dipole is detected in freqmap
    if (
        params["comp"] == "freqmap"
        and params["min"] is None
        and params["max"] is None
        and params["rng"] is None
    ):
        # if colorbar is super-saturated
        while (
            len(params["data"][abs(params["data"]) > params["ticks"][-1]])
            / len(params["data"].ravel())
            > 0.7
        ):
            params["ticks"] = [tick * 10 for tick in params["ticks"]]
            print("[magenta]Colormap saturated. Expanding color-range.[/magenta]")

    # Create ticklabels from final ticks
    params["ticklabels"] = format_list(params["ticks"])

    return params



class ModelComponentNotFoundError(Exception):
    """Raised if a component is missing from the model."""

class ModelComponentError(Exception):
    """Raised if there is an issue with a component in the model."""



import matplotlib.pyplot as plt
import numpy as np

from .plottools import *
from cosmoglobe.h5.chain import Chain


def trace(
    input,
    dataset=None,
    sig=0,
    labels=None,
    showval=True,
    burnin=0,
    xlabel=None,
    ylabel=None,
    nbins=None,
    cmap="tab10",
    figsize=(8, 3),
    darkmode=False,
    fignum=None,
    subplot=None,
    hold=False,
    reuse_axes=False,
):
    if isinstance(input, str) and input.endswith(".h5") and dataset is not None:
        chain = Chain(input)
        if ylabel == None:
            ylabel = dataset
        #component, *items = dataset.split("/")
        input = chain.get(dataset)

    # Make figure
    fig, ax = make_fig(
        figsize,
        fignum,
        hold,
        subplot,
        reuse_axes,
        darkmode,
    )

    if input.ndim < 2:
        input.reshape(-1, 1, 1)
    elif input.ndim == 2:
        input.reshape(0, 1, -1)
    Nsamp, Nsig, Ncomp = input.shape

    cmap = load_cmap(cmap)
    positions = legend_positions(
        input[:, sig, :],
    )

    for i in range(Ncomp):
        plt.plot(
            input[:, sig, i],
            color=cmap(i),
            linewidth=2,
        )

        # Add the text to the right
        if labels is not None:
            hpos = Nsamp * 1.01
            plt.text(
                hpos,
                positions[i],
                rf"{labels[i]}",
                color=cmap(i),
                fontweight="normal",
            )

        if showval:
            mean = np.mean(input[burnin:, sig, i])
            std = np.std(input[burnin:, sig, i])
            label2 = rf"{mean:.2f}$\pm${std:.2f}"
            valpos = Nsamp * 1.01 if labels is None else Nsamp * 1.1
            plt.text(valpos, positions[i], label2, color=cmap(i), fontweight="normal")

    ax.set_xlim(right=Nsamp)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    # Tick and spine parameters
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.tick_params(axis="both", which="both", direction="in")
    plt.yticks(
        rotation=90,
        va="center",
    )
    plt.subplots_adjust(wspace=0, hspace=0.0, right=1)
    plt.tight_layout()
